import numpy as np
import pytest

from xugrid.regrid import reduce


def forward_args():
    values = np.array([0.0, 1.0, 2.0, np.nan])
    weights = np.array([0.5, 0.5, 0.5, 0.5])
    work = np.empty_like(weights)
    return values, weights, work


def reverse_args():
    values, weights, work = forward_args()
    return np.flip(values), weights, work


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_mean(args):
    actual = reduce.mean(*args)
    assert np.allclose(actual, 1.0)


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_harmonic_mean(args):
    actual = reduce.harmonic_mean(*args)
    assert np.allclose(actual, 1.0 / (0.5 / 1.0 + 0.5 / 2.0))


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_geometric_mean(args):
    actual = reduce.geometric_mean(*args)
    assert np.allclose(actual, np.sqrt(1.0 * 2.0))


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_sum(args):
    actual = reduce.sum(*args)
    assert np.allclose(actual, 3.0)


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_minimum(args):
    actual = reduce.minimum(*args)
    assert np.allclose(actual, 0.0)


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_maximum(args):
    actual = reduce.maximum(*args)
    assert np.allclose(actual, 2.0)


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_mode(args):
    actual = reduce.mode(*args)
    # We have tied frequency (all weights 0.5). In this case we return the
    # highest value.
    assert np.allclose(actual, 2.0)


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_median(args):
    actual = reduce.median(*args)
    assert np.allclose(actual, 1.0)


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_conductance(args):
    actual = reduce.conductance(*args)
    assert np.allclose(actual, 1.5)


@pytest.mark.parametrize("args", [forward_args(), reverse_args()])
def test_max_overlap(args):
    actual = reduce.max_overlap(*args)
    # We have tied overlap (all 0.5). In this case we return the highest value.
    assert np.allclose(actual, 2.0)


def test_max_overlap_extra():
    values = np.array([0.0, 1.0, 2.0, np.nan])
    weights = np.array([0.5, 1.5, 0.5, 2.5])
    workspace = np.empty_like(weights)
    args = (values, weights, workspace)
    actual = reduce.max_overlap(*args)
    assert np.allclose(actual, 1.0)


def test_mode_extra():
    values = np.array([0.0, 1.0, 1.0, 2.0, np.nan])
    weights = np.array([0.5, 0.5, 0.5, 0.5, 0.5])
    workspace = np.empty_like(weights)
    args = (values, weights, workspace)
    actual = reduce.mode(*args)
    assert np.allclose(actual, 1.0)
    # The weights shouldn't be mutated!
    assert np.allclose(weights, 0.5)

    values = np.array([1, 1, 3])
    weights = np.array([1.0, 1.0, 1.0])
    workspace = np.empty_like(weights)
    args = (values, weights, workspace)
    actual = reduce.mode(*args)
    assert np.allclose(actual, 1.0)

    values = np.array([4, 5, 6])
    weights = np.array([0.5, 0.5, 0.5])
    workspace = np.empty_like(weights)
    args = (values, weights, workspace)
    actual = reduce.mode(*args)
    # Returns last not-nan value
    assert np.allclose(actual, 6)
    assert np.allclose(weights, 0.5)


def test_percentile():
    # Simplified from:
    # https://github.com/numba/numba/blob/2001717f3321a5082c39c5787676320e699aed12/numba/tests/test_array_reductions.py#L396
    def func(x, p):
        p = np.atleast_1d(p)
        values = x.ravel()
        weights = np.empty_like(values)
        work = np.empty_like(values)
        return np.array([reduce.percentile(values, weights, work, pval) for pval in p])

    q_upper_bound = 100.0
    x = np.arange(8) * 0.5
    np.testing.assert_equal(func(x, 0), 0.0)
    np.testing.assert_equal(func(x, q_upper_bound), 3.5)
    np.testing.assert_equal(func(x, q_upper_bound / 2), 1.75)

    x = np.arange(12).reshape(3, 4)
    q = np.array((0.25, 0.5, 1.0)) * q_upper_bound
    np.testing.assert_equal(func(x, q), [2.75, 5.5, 11.0])

    x = np.arange(3 * 4 * 5 * 6).reshape(3, 4, 5, 6)
    q = np.array((0.25, 0.50)) * q_upper_bound
    np.testing.assert_equal(func(x, q).shape, (2,))

    q = np.array((0.25, 0.50, 0.75)) * q_upper_bound
    np.testing.assert_equal(func(x, q).shape, (3,))

    x = np.arange(12).reshape(3, 4)
    np.testing.assert_equal(func(x, q_upper_bound / 2), 5.5)

    np.testing.assert_equal(func(np.array([1, 2, 3]), 0), 1)

    a = np.array([2, 3, 4, 1])
    func(a, [q_upper_bound / 2])
    np.testing.assert_equal(a, np.array([2, 3, 4, 1]))


METHODS = [
    reduce.mean,
    reduce.harmonic_mean,
    reduce.geometric_mean,
    reduce.sum,
    reduce.minimum,
    reduce.maximum,
    reduce.mode,
    reduce.first_order_conservative,
    reduce.conductance,
    reduce.max_overlap,
    reduce.median,
]


@pytest.mark.parametrize("f", METHODS)
def test_weights_all_zeros(f):
    values = np.ones(5)
    weights = np.zeros(5)
    workspace = np.zeros(5)
    assert np.isnan(f(values, weights, workspace))


@pytest.mark.parametrize("f", METHODS)
def test_values_all_nan(f):
    values = np.full(5, np.nan)
    weights = np.ones(5)
    workspace = np.zeros(5)
    assert np.isnan(f(values, weights, workspace))



import geopandas as gpd
import numpy as np
import pytest
import shapely
import shapely.geometry as sg
import xarray as xr
from pytest_cases import parametrize_with_cases

import xugrid as xu
from xugrid.ugrid.snapping import (
    create_snap_to_grid_dataframe,
    snap_nodes,
    snap_to_grid,
    snap_to_nodes,
)


@pytest.fixture(scope="function")
def structured():
    shape = nlay, nrow, ncol = 3, 9, 9
    dx = 10.0
    dy = -10.0
    xmin = 0.0
    xmax = dx * ncol
    ymin = 0.0
    ymax = abs(dy) * nrow
    dims = ("layer", "y", "x")

    layer = np.arange(1, nlay + 1)
    y = np.arange(ymax, ymin, dy) + 0.5 * dy
    x = np.arange(xmin, xmax, dx) + 0.5 * dx
    coords = {"layer": layer, "y": y, "x": x}

    return xr.DataArray(np.ones(shape, dtype=np.int32), coords=coords, dims=dims)


def test_snap__three_points_horizontal():
    x = np.array([0.0, 1.0, 2.0])
    y = np.zeros_like(x)
    inv_perm, snap_x, snap_y = snap_nodes(x, y, 0.1)
    assert inv_perm is None
    assert np.array_equal(x, snap_x)
    assert np.array_equal(y, snap_y)

    inv_perm, snap_x, snap_y = snap_nodes(x, y, 1.0)
    expected_inv_perm = np.array([0, 0, 1])
    expected_x = np.array([0.0, 2.0])
    expected_y = np.zeros_like(expected_x)
    assert np.array_equal(inv_perm, expected_inv_perm)
    assert np.array_equal(expected_x, snap_x)
    assert np.array_equal(expected_y, snap_y)

    inv_perm, snap_x, snap_y = snap_nodes(x, y, 2.0)
    expected_inv_perm = np.array([0, 0, 0])
    expected_x = np.array([0.0])
    expected_y = np.zeros_like(expected_x)
    assert np.array_equal(inv_perm, expected_inv_perm)
    assert np.array_equal(expected_x, snap_x)
    assert np.array_equal(expected_y, snap_y)


def test_snap__three_points_diagonal():
    x = y = np.array([0.0, 1.0, 1.5])
    inv_perm, snap_x, snap_y = snap_nodes(x, y, 0.1)
    assert inv_perm is None
    assert np.array_equal(x, snap_x)
    assert np.array_equal(y, snap_y)

    # hypot(0.5, 0.5) = 0.707...
    inv_perm, snap_x, snap_y = snap_nodes(x, y, 0.71)
    expected_inv_perm = np.array([0, 1, 1])
    expected_x = expected_y = np.array([0.0, 1.0])
    assert np.array_equal(inv_perm, expected_inv_perm)
    assert np.array_equal(snap_x, expected_x)
    assert np.array_equal(snap_y, expected_y)

    # hypot(1, 1) = 1.414...
    inv_perm, snap_x, snap_y = snap_nodes(x, y, 1.42)
    expected_inv_perm = np.array([0, 1, 1])
    expected_x = expected_y = np.array([0.0, 1.5])
    assert np.array_equal(inv_perm, expected_inv_perm)
    assert np.array_equal(snap_x, expected_x)
    assert np.array_equal(snap_y, expected_y)


def test_snap__two_lines():
    x = np.array([0.0, 1.0, 1.02, 2.0])
    y = np.array([1.0, 0.0, 0.0, 1.0])
    edge_node_connectivity = np.array(
        [
            [0, 1],
            [2, 3],
        ]
    )
    inv_perm, snap_x, snap_y = snap_nodes(x, y, 0.1)
    c = inv_perm[edge_node_connectivity]

    expected_inv_perm = np.array([0, 1, 1, 2])
    expected_x = np.array([0.0, 1.0, 2.0])
    expected_y = np.array([1.0, 0.0, 1.0])
    expected_c = np.array(
        [
            [0, 1],
            [1, 2],
        ]
    )
    assert np.array_equal(inv_perm, expected_inv_perm)
    assert np.array_equal(snap_x, expected_x)
    assert np.array_equal(snap_y, expected_y)
    assert np.array_equal(c, expected_c)


def test_snap_to_nodes():
    x = np.array([1.0, 2.0, 3.0])
    y = np.array([1.0, 2.0, 3.0])
    to_x = x + 0.1
    to_y = y + 0.1

    # None snapped
    snap_x, snap_y = snap_to_nodes(x, y, to_x, to_y, 0.1)
    assert np.array_equal(snap_x, x)
    assert np.array_equal(snap_y, y)

    # All snapped
    snap_x, snap_y = snap_to_nodes(x, y, to_x, to_y, 0.2)
    assert np.array_equal(snap_x, to_x)
    assert np.array_equal(snap_y, to_y)

    # Ties, no tiebreaker
    with pytest.raises(ValueError):
        snap_x, snap_y = snap_to_nodes(x, y, to_x, to_y, 3.0)

    # Take nearest
    snap_x, snap_y = snap_to_nodes(x, y, to_x, to_y, 3.0, tiebreaker="nearest")
    assert np.array_equal(snap_x, to_x)
    assert np.array_equal(snap_y, to_y)

    # More ties
    to_x = np.array([1.01, 2.01, 2.002, 3.01])
    to_y = np.array([1.01, 2.01, 2.002, 3.01])
    snap_x, snap_y = snap_to_nodes(x, y, to_x, to_y, 0.5, tiebreaker="nearest")
    expected_x = np.array([1.01, 2.002, 3.01])
    expected_y = np.array([1.01, 2.002, 3.01])
    assert np.array_equal(snap_x, expected_x)
    assert np.array_equal(snap_y, expected_y)

    # Exact ties
    to_x = np.array([1.01, 2.002, 2.002, 3.01])
    to_y = np.array([1.01, 2.002, 2.002, 3.01])
    snap_x, snap_y = snap_to_nodes(x, y, to_x, to_y, 0.5, tiebreaker="nearest")
    assert np.array_equal(snap_x, expected_x)
    assert np.array_equal(snap_y, expected_y)

    # Multiple ties
    to_x = np.array([1.01, 2.01, 2.002, 3.002, 3.01])
    to_y = np.array([1.01, 2.01, 2.002, 3.002, 3.01])
    snap_x, snap_y = snap_to_nodes(x, y, to_x, to_y, 0.5, tiebreaker="nearest")
    expected_x = np.array([1.01, 2.002, 3.002])
    expected_y = np.array([1.01, 2.002, 3.002])
    assert np.array_equal(snap_x, expected_x)
    assert np.array_equal(snap_y, expected_y)


def test_snap_to_grid():
    idomain = xr.DataArray(
        data=[[1, 1], [1, 1]],
        coords={"y": [0.5, 1.5], "x": [0.5, 1.5]},
        dims=["y", "x"],
    )
    line = sg.LineString([[0.5, 0.0], [1.5, 2.0]])
    line_gdf = gpd.GeoDataFrame({"resistance": [100.0]}, geometry=[line])
    uds, gdf = snap_to_grid(line_gdf, idomain, 2.0)
    assert isinstance(uds, xu.UgridDataset)
    assert isinstance(gdf, gpd.GeoDataFrame)
    assert uds["resistance"].dims == (uds.ugrid.grid.edge_dimension,)
    # TODO test for returned values...


class LineCases:
    def case_single_line(self):
        line_x = [40.2, 40.2, 40.2]
        line_y = [82.0, 40.0, 0.0]
        geometry = gpd.GeoDataFrame(
            geometry=[shapely.linestrings(line_x, line_y)], data={"a": [1.0]}
        )
        unique_values = np.array([0.0, np.nan])
        line_counts = np.array([8, 172])
        return geometry, unique_values, line_counts

    def case_single_line_at_edge(self):
        line_x = [40.0, 40.0, 40.0]
        line_y = [82.0, 40.0, 0.0]
        geometry = gpd.GeoDataFrame(
            geometry=[shapely.linestrings(line_x, line_y)], data={"a": [1.0]}
        )
        unique_values = np.array([0.0, np.nan])
        line_counts = np.array([8, 172])
        return geometry, unique_values, line_counts

    def case_parallel_lines(self):
        line_x1 = [10.2, 10.2, 10.2]
        line_x2 = [30.2, 30.2, 30.2]
        line_y = [82.0, 40.0, 0.0]
        line1 = shapely.linestrings(line_x1, line_y)
        line2 = shapely.linestrings(line_x2, line_y)
        geometry = gpd.GeoDataFrame(geometry=[line1, line2], data={"a": [1.0, 1.0]})

        unique_values = np.array([0.0, 1.0, np.nan])
        line_counts = np.array([8, 8, 164])
        return geometry, unique_values, line_counts

    def case_series_lines(self):
        # This caused a failure up to 0.10.0
        line_x = [40.2, 40.2]
        line_y1 = [82.0, 60.0]
        line_y2 = [60.0, 40.0]
        line_y3 = [40.0, 20.0]
        line_y4 = [20.0, 0.0]
        line1 = shapely.linestrings(line_x, line_y1)
        line2 = shapely.linestrings(line_x, line_y2)
        line3 = shapely.linestrings(line_x, line_y3)
        line4 = shapely.linestrings(line_x, line_y4)
        geometry = gpd.GeoDataFrame(
            geometry=[line1, line2, line3, line4], data={"a": [1.0, 1.0, 1.0, 1.0]}
        )
        unique_values = np.array([0.0, 1.0, 2.0, 3.0, np.nan])
        line_counts = np.array([2, 2, 2, 2, 172])
        return geometry, unique_values, line_counts

    def case_crossing_lines(self):
        # This caused a failure up to 0.10.0
        line_x = [40.2, 40.2, 40.2]
        line_y = [82.0, 40.0, 0.0]
        line1 = shapely.linestrings(line_x, line_y)
        line2 = shapely.linestrings(line_y, line_x)
        geometry = gpd.GeoDataFrame(geometry=[line1, line2], data={"a": [1.0, 2.0]})

        unique_values = np.array([0.0, 1.0, np.nan])
        line_counts = np.array([8, 8, 164])
        return geometry, unique_values, line_counts

    def case_closely_parallel(self):
        """
        Snap closely parallel lines. These are snapped to same edge, the first
        one should be taken. We can use this test to monitor if this behaviour
        changes.
        """
        line_x1 = [19.0, 19.0, 19.0]
        line_x2 = [21.0, 21.0, 21.0]
        line_y = [82.0, 40.0, 0.0]
        line1 = shapely.linestrings(line_x1, line_y)
        line2 = shapely.linestrings(line_x2, line_y)
        geometry = gpd.GeoDataFrame(geometry=[line1, line2], data={"a": [1.0, 1.0]})

        unique_values = np.array([0.0, np.nan])
        line_counts = np.array([8, 172])
        return geometry, unique_values, line_counts

    def case_line_hits_edge_centroid(self):
        """
        Snap a line that hits the centroid of an edge. This is a special case
        that caused problems with the snapping algorithm up to xugrid 0.14.1:
        the line exactly hits the intersection point of the line connecting the
        face centroids and the cell edge.
        """
        line_x = [12.0, 18.0]
        line_y = [22.0, 18.0]
        line = shapely.linestrings(line_x, line_y)
        geometry = gpd.GeoDataFrame(geometry=[line], data={"a": [1.0]})

        unique_values = np.array([0.0, np.nan])
        line_counts = np.array([1, 179])
        return geometry, unique_values, line_counts


@parametrize_with_cases(["geometry", "unique_values", "line_counts"], cases=LineCases)
def test_snap_to_grid_with_data(structured, geometry, unique_values, line_counts):
    uds, gdf = snap_to_grid(geometry, structured, max_snap_distance=0.5)
    assert isinstance(uds, xu.UgridDataset)
    assert isinstance(gdf, gpd.GeoDataFrame)
    assert uds["a"].dims == (uds.ugrid.grid.edge_dimension,)

    actual_unique_values, actual_line_counts = np.unique(
        uds["line_index"], return_counts=True
    )
    np.testing.assert_array_equal(unique_values, actual_unique_values)
    np.testing.assert_array_equal(line_counts, actual_line_counts)


def test_create_snap_to_grid_dataframe(structured):
    """
    Test if create_snap_to_grid_dataframe and its code example in the docstring
    still work
    """
    # Arrange test
    line_x1 = [19.0, 19.0, 19.0]
    line_x2 = [21.0, 21.0, 21.0]
    line_y = [82.0, 40.0, 0.0]

    line1 = shapely.linestrings(line_x1, line_y)
    line2 = shapely.linestrings(line_x2, line_y)
    lines = gpd.GeoDataFrame(geometry=[line1, line2], data={"my_variable": [1.0, 2.0]})
    grid2d = xu.Ugrid2d.from_structured(structured)
    edge_arr_data = np.ones(grid2d.sizes["mesh2d_nEdges"])
    edge_data = xu.UgridDataArray.from_data(edge_arr_data, grid2d, facet="edge")

    # Act
    snapping_df = create_snap_to_grid_dataframe(lines, grid2d, max_snap_distance=0.5)

    # Assert
    edge_index0 = snapping_df.loc[
        snapping_df["line_index"] == 0, "edge_index"
    ].to_numpy()
    edge_index1 = snapping_df.loc[
        snapping_df["line_index"] == 1, "edge_index"
    ].to_numpy()
    assert np.all(edge_index0 == edge_index1)

    # Now test if docstring part works
    snapping_df["my_variable"] = (
        lines["my_variable"].iloc[snapping_df["line_index"]].to_numpy()
    )
    aggregated = snapping_df.groupby("edge_index").sum()

    new = xu.full_like(edge_data, np.nan)
    new.data[aggregated.index] = aggregated["my_variable"]

    # A lot of values are NaN, but all not-nan values should be three.
    np.testing.assert_array_equal(np.unique(new), [3, np.nan])



import numpy as np
import pytest

import xugrid as xu
from xugrid.ugrid import connectivity, voronoi


def rowsort(a):
    return a[np.lexsort(a.T)]


def polygon_area(p):
    area = 0.0
    n = len(p)
    for i in range(n):
        v0 = p[i]
        v1 = p[(i + 1) % n]
        area += v0[0] * v1[1]
        area -= v0[1] * v1[0]
    return area


def mesh_area(vertices, faces):
    area_sum = 0.0
    for face in faces:
        polygon = vertices[face[face != -1]]
        area_sum += polygon_area(polygon)
    return 0.5 * abs(area_sum)


def test_dot_product2d():
    U = np.array([[1.0, 2.0], [3.0, 4.0]])
    V = np.array([[5.0, 6.0], [7.0, 8.0]])
    assert np.allclose(voronoi.dot_product2d(U, V), [17.0, 53.0])


def test_compute_centroid():
    x = np.array([0.0, 1.0, 1.0, 0.0, 1.0, 2.0, 2.0, 1.0])
    y = np.array([0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0])
    i = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    expected_x = np.array([0.5, 1.5])
    expected_y = np.array([0.5, 1.5])

    actual_x, actual_y = voronoi._centroid_pandas(i, x, y)
    assert np.allclose(actual_x, expected_x)
    assert np.allclose(actual_y, expected_y)

    actual_x, actual_y = voronoi._centroid_scipy(i, x, y)
    assert np.allclose(actual_x, expected_x)
    assert np.allclose(actual_y, expected_y)

    actual_x, actual_y = voronoi.compute_centroid(i, x, y)
    assert np.allclose(actual_x, expected_x)
    assert np.allclose(actual_y, expected_y)


class TestVoronoi:
    """
    Note: Arguably the best way to check these tests is by plotting the
    resulting polygons.
    """

    @pytest.fixture(autouse=True)
    def setup(self):
        # Make a simple rectangular mesh of 2 rows, 3 columns
        self.vertices = np.array(
            [
                [0.0, 0.0],  # 0
                [1.0, 0.0],  # 1
                [2.0, 0.0],  # 2
                [3.0, 0.0],  # 3
                [0.0, 1.0],  # 4
                [1.0, 1.0],  # 5
                [2.0, 1.0],  # 6
                [3.0, 1.0],  # 7
                [0.0, 2.0],  # 8
                [1.0, 2.0],  # 9
                [2.0, 2.0],  # 10
                [3.0, 2.0],  # 11
            ]
        )
        self.face_node_connectivity = np.array(
            [
                [0, 1, 5, 4],
                [1, 2, 6, 5],
                [2, 3, 7, 6],
                [4, 5, 9, 8],
                [5, 6, 10, 9],
                [6, 7, 11, 10],
            ]
        )
        self.node_face_connectivity = connectivity.invert_dense_to_sparse(
            self.face_node_connectivity
        )
        (
            self.edge_node_connectivity,
            face_edge_connectivity,
        ) = connectivity.edge_connectivity(
            self.face_node_connectivity,
        )
        self.edge_face_connectivity = connectivity.invert_dense(face_edge_connectivity)
        self.centroids = np.array(
            [
                [0.5, 0.5],
                [1.5, 0.5],
                [2.5, 0.5],
                [0.5, 1.5],
                [1.5, 1.5],
                [2.5, 1.5],
            ]
        )
        self.expected_vertices = rowsort(self.centroids)

        self.exterior_vertices = rowsort(
            np.array(
                [
                    [0.0, 0.5],  # left
                    [0.0, 1.5],  # left
                    [3.0, 0.5],  # right
                    [3.0, 1.5],  # right
                    [0.5, 0.0],  # bottom
                    [1.5, 0.0],  # bottom
                    [2.5, 0.0],  # bottom
                    [0.5, 2.0],  # top
                    [1.5, 2.0],  # top
                    [2.5, 2.0],  # top
                ]
            )
        )
        self.additional_vertices = rowsort(
            np.array(
                [
                    [0.0, 0.0],  # 0
                    [1.0, 0.0],  # 1
                    [2.0, 0.0],  # 2
                    [3.0, 0.0],  # 3
                    [0.0, 1.0],  # 4
                    [3.0, 1.0],  # 7
                    [0.0, 2.0],  # 8
                    [1.0, 2.0],  # 9
                    [2.0, 2.0],  # 10
                    [3.0, 2.0],  # 11
                ]
            )
        )

    def test_exterior_centroids(self):
        actual_i, actual_j = voronoi.exterior_centroids(self.node_face_connectivity)
        expected_i = np.array([0, 3, 8, 11])
        expected_j = np.array([0, 2, 3, 5])
        assert np.array_equal(actual_i, expected_i)
        assert np.array_equal(actual_j, expected_j)

    def test_interior_centroids(self):
        actual_i, actual_j = voronoi.interior_centroids(
            self.node_face_connectivity,
            self.edge_face_connectivity,
            self.edge_node_connectivity,
        )
        expected_i = np.array([1, 1, 2, 2, 4, 4, 7, 7, 9, 9, 10, 10])
        expected_j = np.array([0, 1, 1, 2, 0, 3, 2, 5, 3, 4, 4, 5])
        assert np.array_equal(actual_i, expected_i)
        assert np.array_equal(actual_j, expected_j)

    def test_exterior_vertices(self):
        (
            _,
            _,
            actual_vertices,
            actual_face,
            n,
            interpolation_map,
        ) = voronoi.exterior_vertices(
            self.edge_face_connectivity,
            self.edge_node_connectivity,
            self.vertices,
            self.centroids,
            add_vertices=False,
        )
        assert n == 0
        assert np.allclose(rowsort(actual_vertices), self.exterior_vertices)
        assert np.isin(np.arange(6), actual_face).all()
        assert interpolation_map is None

    def test_voronoi_topology(self):
        vertices, faces, face_i, _ = voronoi.voronoi_topology(
            self.node_face_connectivity,
            self.vertices,
            self.centroids,
        )
        expected_faces = np.array(
            [
                [0, 1, 4, 3],
                [1, 2, 5, 4],
            ]
        )
        assert faces.shape == (2, 4)
        assert np.allclose(rowsort(vertices), self.expected_vertices)
        assert np.array_equal(face_i, [0, 1, 2, 3, 4, 5])
        assert np.array_equal(faces, expected_faces)
        assert np.allclose(mesh_area(vertices, faces), 2.0)

    def test_voronoi_topology__add_exterior(self):
        with pytest.raises(
            ValueError, match="must be provided if add_exterior is True"
        ):
            voronoi.voronoi_topology(
                self.node_face_connectivity,
                self.vertices,
                self.centroids,
                add_exterior=True,
            )

        vertices, faces, face_i, _ = voronoi.voronoi_topology(
            self.node_face_connectivity,
            self.vertices,
            self.centroids,
            self.edge_face_connectivity,
            self.edge_node_connectivity,
            add_exterior=True,
        )
        expected_vertices = rowsort(
            np.concatenate([self.expected_vertices, self.exterior_vertices])
        )
        assert faces.shape == (12, 4)
        assert np.allclose(rowsort(vertices), expected_vertices)
        assert (face_i != -1).all()
        assert np.allclose(mesh_area(vertices, faces), 5.5)

        vertices, faces, face_i, node_interpolation = voronoi.voronoi_topology(
            self.node_face_connectivity,
            self.vertices,
            self.centroids,
            self.edge_face_connectivity,
            self.edge_node_connectivity,
            add_exterior=True,
            add_vertices=True,
        )
        expected_vertices = rowsort(
            np.concatenate([expected_vertices, self.additional_vertices])
        )
        # This introduces hanging nodes
        assert faces.shape == (12, 5)
        assert np.allclose(rowsort(vertices), expected_vertices)
        assert (face_i == -1).sum() == 10
        assert np.allclose(mesh_area(vertices, faces), 6.0)


def test_projected_vertices_on_edge():
    """
    For certain triangles, the voronoi projection falls exactly on the edge.
         x
        ---
      -------
    x -- o -- x
         |
         |
         v

    Where:

    * x: the triangle vertices
    * o: the circumcenter
    * -->: the orthogonal voronoi ray

    This results in a centroid which is identical to the voronoi ray edge
    intersection. This will then create a zero length edge, which is obviously
    problematic.
    """
    nodes = np.array(
        [
            [0.0, 0.0],  # 0
            [0.0, 2.0],  # 1
            [2.0, 2.0],  # 2
            [0.0, 2.0],  # 3
            [1.0, 1.0],  # 4
        ]
    )
    faces = np.array(
        [
            [0, 1, 4],
            [1, 2, 4],
            [2, 3, 4],
            [3, 0, 4],
        ]
    )
    grid = xu.Ugrid2d(nodes[:, 0], nodes[:, 1], -1, faces)
    voronoi_grid = grid.tesselate_circumcenter_voronoi()
    assert voronoi_grid.n_face == 5


def test_isolated_face():
    """
    The centroid of the left-most face should not be included.

          +--+
          |  |
    +--+--+--+
    |  |  |  |
    +--+--+--+
          |  |
          +--+

    """
    nodes = np.array(
        [
            [0.0, 0.0],  # 0
            [1.0, 0.0],  # 1
            [2.0, 0.0],  # 2
            [3.0, 0.0],  # 3
            [0.0, 1.0],  # 4
            [1.0, 1.0],  # 5
            [2.0, 1.0],  # 6
            [3.0, 1.0],  # 7
            [2.0, -1.0],  # 8
            [3.0, -1.0],  # 9
            [2.0, 2.0],  # 10
            [3.0, 2.0],  # 11
        ]
    )
    faces = np.array(
        [
            [0, 1, 5, 4],
            [1, 2, 6, 5],
            [2, 3, 7, 6],
            [8, 9, 3, 2],
            [6, 7, 11, 10],
        ]
    )
    grid = xu.Ugrid2d(nodes[:, 0], nodes[:, 1], -1, faces)
    voronoi_grid = grid.tesselate_centroidal_voronoi(False, False)
    assert voronoi_grid.n_face == 2
    assert voronoi_grid.n_node == 4


def test_concave_voronoi():
    r"""
    Three triangles.
    The lack of a fourth on the creates

    4 (0,2)      5 (3,2)
         *------------*
         | \   2    / |
         |   \    /   |
         |  1  \/     |
         |     /\     |
         |   /    \   |
         | /   0    \ |
         *------------*
    1 (0,0)  3 (1,1)  2 (3,0)
    """
    vertices = np.array(
        [
            [0.0, 0.0],
            [3.0, 0.0],
            [1.0, 1.0],
            [0.0, 2.0],
            [3.0, 2.0],
        ]
    )
    faces = np.array(
        [
            [0, 1, 2],
            [0, 2, 3],
            [2, 4, 3],
        ]
    )
    grid = xu.Ugrid2d(*vertices.T, -1, faces)
    voronoi_0 = grid.tesselate_centroidal_voronoi(skip_concave=False)
    voronoi_1 = grid.tesselate_centroidal_voronoi(skip_concave=True)
    assert voronoi_0.n_face == voronoi_1.n_face
    assert voronoi_0.n_node == voronoi_1.n_node
    assert voronoi_0.area.sum() < voronoi_1.area.sum()



import numpy as np
import pytest

import xugrid
from xugrid.regrid.unstructured import UnstructuredGrid2d


@pytest.fixture(scope="function")
def circle():
    return UnstructuredGrid2d(xugrid.data.disk().grid)


def test_init():
    uds = xugrid.data.disk()
    assert isinstance(UnstructuredGrid2d(uds), UnstructuredGrid2d)
    assert isinstance(UnstructuredGrid2d(uds["face_z"]), UnstructuredGrid2d)
    assert isinstance(UnstructuredGrid2d(uds.ugrid.grid), UnstructuredGrid2d)
    with pytest.raises(TypeError):
        UnstructuredGrid2d(1)


def test_grid_properties(circle):
    assert circle.dims == ("mesh2d_nFaces",)
    assert circle.shape == (384,)
    assert circle.size == 384
    assert isinstance(circle.area, np.ndarray)
    assert circle.area.size == 384


@pytest.mark.parametrize("relative", [True, False])
def test_overlap(circle, relative):
    source, target, weights = circle.overlap(other=circle, relative=relative)
    valid = weights > 1.0e-5
    source = source[valid]
    target = target[valid]
    weights = weights[valid]
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.arange(circle.size))
    assert np.array_equal(target[sorter], np.arange(circle.size))
    if relative:
        assert np.allclose(weights[sorter], np.ones(circle.size))
    else:
        assert np.allclose(weights[sorter], circle.area)


def test_locate_centroids(circle):
    source, target, weights = circle.locate_centroids(circle)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.arange(circle.size))
    assert np.array_equal(target[sorter], np.arange(circle.size))
    assert np.allclose(weights[sorter], np.ones(circle.size))


def test_barycentric(circle):
    source, target, weights = circle.barycentric(circle)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.arange(circle.size))
    assert np.array_equal(target[sorter], np.arange(circle.size))
    assert np.allclose(weights[sorter], np.ones(circle.size))



from xugrid.plot.plot import (
    contour,
    contourf,
    imshow,
    line,
    pcolormesh,
    scatter,
    surface,
    tripcolor,
)

__all__ = (
    "contour",
    "contourf",
    "imshow",
    "line",
    "pcolormesh",
    "scatter",
    "surface",
    "tripcolor",
)



from functools import wraps

import xarray as xr

from xugrid.core.utils import unique_grids
from xugrid.core.wrap import UgridDataArray, UgridDataset

DATAARRAY_NAME = "__xarray_dataarray_name__"
DATAARRAY_VARIABLE = "__xarray_dataarray_variable__"


def _dataset_helper(ds: xr.Dataset):
    n_topology = len(ds.ugrid_roles.topology)
    if n_topology == 0:
        raise ValueError(
            "The file or object does not contain UGRID conventions data. "
            "One or more UGRID topologies are required. Perhaps you wrote "
            "the file using `data.to_netcdf()` instead of `data.ugrid.to_netcdf()`?"
        )
    return UgridDataset(ds)


def open_dataset(*args, **kwargs):
    ds = xr.open_dataset(*args, **kwargs)
    return _dataset_helper(ds)


def load_dataset(*args, **kwargs):
    ds = xr.load_dataset(*args, **kwargs)
    return _dataset_helper(ds)


def _dataarray_helper(ds: xr.Dataset):
    dataset = _dataset_helper(ds)
    if len(dataset.data_vars) != 1:
        raise ValueError(
            "The file or object contains more than one data "
            "variable. Please read with xarray.open_dataset and "
            "then select the variable you want."
        )
    else:
        (data_array,) = dataset.data_vars.values()

    data_array.set_close(dataset._close)

    # Reset names if they were changed during saving
    # to ensure that we can 'roundtrip' perfectly
    if DATAARRAY_NAME in dataset.attrs:
        data_array.name = dataset.attrs[DATAARRAY_NAME]
        del dataset.attrs[DATAARRAY_NAME]

    if data_array.name == DATAARRAY_VARIABLE:
        data_array.name = None

    return UgridDataArray(data_array, dataset.grid)


def load_dataarray(*args, **kwargs):
    ds = xr.load_dataset(*args, **kwargs)
    return _dataarray_helper(ds)


def open_dataarray(*args, **kwargs):
    ds = xr.open_dataset(*args, **kwargs)
    return _dataarray_helper(ds)


def open_mfdataset(*args, **kwargs):
    if "data_vars" in kwargs:
        raise ValueError("data_vars kwargs is not supported in xugrid.open_mfdataset")
    kwargs["data_vars"] = "minimal"
    ds = xr.open_mfdataset(*args, **kwargs)
    return UgridDataset(ds)


def open_zarr(*args, **kwargs):
    ds = xr.open_zarr(*args, **kwargs)
    return UgridDataset(ds)


load_dataset.__doc__ = xr.load_dataset.__doc__
open_dataset.__doc__ = xr.open_dataset.__doc__
load_dataarray.__doc__ = xr.load_dataarray.__doc__
open_dataarray.__doc__ = xr.open_dataarray.__doc__
open_mfdataset.__doc__ = xr.open_mfdataset.__doc__
open_zarr.__doc__ = xr.open_zarr.__doc__


# Other utilities
# ---------------


def wrap_func_like(func):
    @wraps(func)
    def _like(other, *args, **kwargs):
        obj = func(other.obj, *args, **kwargs)
        if isinstance(obj, xr.DataArray):
            return type(other)(obj, other.grid)
        elif isinstance(obj, xr.Dataset):
            return type(other)(obj, other.grids)
        else:
            raise TypeError(
                f"Expected Dataset or DataArray, received {type(other).__name__}"
            )

    _like.__doc__ = func.__doc__
    return _like


def wrap_func_objects(func):
    @wraps(func)
    def _f(objects, *args, **kwargs):
        grids = []
        bare_objs = []
        for obj in objects:
            if isinstance(obj, UgridDataArray):
                grids.append(obj.grid)
            elif isinstance(obj, UgridDataset):
                grids.extend(obj.grids)
            else:
                raise TypeError(
                    "Can only concatenate xugrid UgridDataset and UgridDataArray "
                    f"objects, got {type(obj).__name__}"
                )

            bare_objs.append(obj.obj)

        grids = unique_grids(grids)
        result = func(bare_objs, *args, **kwargs)
        if isinstance(result, xr.DataArray):
            if len(grids) > 1:
                raise ValueError("All UgridDataArrays must have the same grid")
            return UgridDataArray(result, next(iter(grids)))
        else:
            return UgridDataset(result, grids)

    _f.__doc__ = func.__doc__
    return _f


full_like = wrap_func_like(xr.full_like)
zeros_like = wrap_func_like(xr.zeros_like)
ones_like = wrap_func_like(xr.ones_like)

concat = wrap_func_objects(xr.concat)
merge = wrap_func_objects(xr.merge)



"""
Copied from xarray.tests.test_utils.py

The reason is that the content of xarray.core.utils are all private methods.
Hence, Xarray provides no guarantees on breaking changes.

Xarray is licensed under Apache License 2.0:
https://github.com/pydata/xarray/blob/main/LICENSE
"""

import numpy as np
import pytest

import xugrid
from xugrid.core.utils import either_dict_or_kwargs, unique_grids


def grid1d(dataset=None, indexes=None, crs=None, attrs=None):
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 2.0],
        ]
    )
    grid = xugrid.Ugrid1d(
        node_x=xy[:, 0],
        node_y=xy[:, 1],
        fill_value=-1,
        edge_node_connectivity=np.array([[0, 1], [1, 2]]),
        dataset=dataset,
        indexes=indexes,
        crs=crs,
        attrs=attrs,
    )
    return grid


def test_either_dict_or_kwargs():
    result = either_dict_or_kwargs({"a": 1}, None, "foo")
    expected = {"a": 1}
    assert result == expected

    result = either_dict_or_kwargs(None, {"a": 1}, "foo")
    expected = {"a": 1}
    assert result == expected

    with pytest.raises(ValueError, match=r"foo"):
        result = either_dict_or_kwargs({"a": 1}, {"a": 1}, "foo")


def test_unique_grids():
    grid = grid1d()
    grid2 = grid1d()
    grid_different = grid1d()

    grid_different._attrs["something"] = "different"

    assert len(unique_grids([grid, grid2, grid_different])) == 2
    assert len(unique_grids([grid, grid2])) == 1
    assert len(unique_grids([grid, grid_different])) == 2



from collections import ChainMap

import numpy as np
import pyproj
import pytest
import xarray as xr

import xugrid
from xugrid.ugrid import conventions as cv


def test_infer_xy_coords():
    def assign_node_coords(ds: xr.Dataset, name: str, stdname: str = None):
        if stdname:
            attrs = {"standard_name": stdname}
        else:
            attrs = None

        da = xr.DataArray([0.0, 1.0], dims=["node"], attrs=attrs)
        return ds.assign_coords({name: da})

    # All have standard names: take all
    ds = xr.Dataset()
    ds = assign_node_coords(ds, "x", "projection_x_coordinate")
    ds = assign_node_coords(ds, "y", "projection_y_coordinate")
    ds = assign_node_coords(ds, "lon", "longitude")
    ds = assign_node_coords(ds, "lat", "latitude")
    candidates = ["x", "y", "lon", "lat"]
    x, y = cv._infer_xy_coords(ds, candidates)
    assert x == ["x", "lon"]
    assert y == ["y", "lat"]

    # Only lon and lat have standard names: take them.
    ds = xr.Dataset()
    ds = assign_node_coords(ds, "x")
    ds = assign_node_coords(ds, "y")
    ds = assign_node_coords(ds, "lon", "longitude")
    ds = assign_node_coords(ds, "lat", "latitude")
    x, y = cv._infer_xy_coords(ds, candidates)
    assert x == ["lon"]
    assert y == ["lat"]

    # Non have standard names: take the first two.
    ds = xr.Dataset()
    ds = assign_node_coords(ds, "x")
    ds = assign_node_coords(ds, "y")
    ds = assign_node_coords(ds, "lon")
    ds = assign_node_coords(ds, "lat")
    candidates = ["x", "y", "lon", "lat"]
    with pytest.warns(UserWarning):
        x, y = cv._infer_xy_coords(ds, candidates)
    assert x == ["x"]
    assert y == ["y"]

    # Only one: error
    ds = xr.Dataset()
    ds = assign_node_coords(ds, "x", "projection_x_coordinate")
    ds = assign_node_coords(ds, "y")
    candidates = ["x", "y"]
    with pytest.raises(cv.UgridCoordinateError):
        x, y = cv._infer_xy_coords(ds, candidates)

    # Only one: error
    ds = xr.Dataset()
    ds = assign_node_coords(ds, "x")
    ds = assign_node_coords(ds, "y", "projection_y_coordinate")
    candidates = ["x", "y"]
    with pytest.raises(cv.UgridCoordinateError):
        x, y = cv._infer_xy_coords(ds, candidates)


class TestConventionsElevation:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.ds = xugrid.data.elevation_nl(xarray=True)
        self.coordinates = {
            "mesh2d": {
                "node_coordinates": (
                    ["mesh2d_node_x"],
                    ["mesh2d_node_y"],
                ),
                "face_coordinates": (
                    ["mesh2d_face_x"],
                    ["mesh2d_face_y"],
                ),
            }
        }
        self.connectivity = {
            "mesh2d": {
                "face_node_connectivity": "mesh2d_face_nodes",
            }
        }
        self.dimensions = {
            "mesh2d": {
                "edge_dimension": "mesh2d_nEdges",
                "face_dimension": "mesh2d_nFaces",
                "node_dimension": "mesh2d_nNodes",
            },
        }

    def test_get_topology(self):
        assert cv._get_topology(self.ds) == ["mesh2d"]

    def test_get_coordinates(self):
        ds = xugrid.data.elevation_nl(xarray=True)
        actual = cv._get_coordinates(ds, ["mesh2d"])
        assert actual == self.coordinates

        ds = xugrid.data.elevation_nl(xarray=True)
        ds["mesh2d"].attrs["edge_coordinates"] = "mesh2d_edge_x mesh2d_edge_y"
        with pytest.warns(UserWarning):
            cv._get_coordinates(ds, ["mesh2d"])

        ds = xugrid.data.elevation_nl(xarray=True)
        ds["mesh2d"].attrs["edge_coordinates"] = "mesh2d_edge_x"
        ds["mesh2d_edge_x"] = 0  # Put a dummy value in the dataset
        with pytest.raises(cv.UgridCoordinateError):
            cv._get_coordinates(ds, ["mesh2d"])

    def test_get_connectivity(self):
        ds = xugrid.data.elevation_nl(xarray=True)
        actual = cv._get_connectivity(ds, ["mesh2d"])
        assert actual == self.connectivity

    def test_get_dimensions(self):
        ds = xugrid.data.elevation_nl(xarray=True)
        connectivity = cv._get_connectivity(ds, ["mesh2d"])
        coordinates = cv._get_coordinates(ds, ["mesh2d"])
        actual = cv._get_dimensions(ds, ["mesh2d"], connectivity, coordinates)
        assert actual == self.dimensions

    def test_topology(self):
        assert self.ds.ugrid_roles.topology == ["mesh2d"]

    def test_coordinates(self):
        assert self.ds.ugrid_roles.coordinates == self.coordinates

    def test_dimensions(self):
        assert self.ds.ugrid_roles.dimensions == self.dimensions

    def test_connectivity(self):
        assert self.ds.ugrid_roles.connectivity == self.connectivity

    def test_getitem(self):
        result = self.ds.ugrid_roles["mesh2d"]
        assert isinstance(result, ChainMap)

        with pytest.raises(KeyError):
            self.ds.ugrid_roles["mesh1d"]

        assert self.ds.ugrid_roles["mesh2d"]["node_coordinates"] == (
            ["mesh2d_node_x"],
            ["mesh2d_node_y"],
        )

    def test_repr(self):
        result = self.ds.ugrid_roles.__repr__()
        assert isinstance(result, str)


class TestCrsConventions:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.ds = xugrid.data.elevation_nl(xarray=True)
        self.coordinates = {
            "mesh2d": {
                "node_coordinates": (
                    ["mesh2d_node_x"],
                    ["mesh2d_node_y"],
                ),
                "face_coordinates": (
                    ["mesh2d_face_x"],
                    ["mesh2d_face_y"],
                ),
            }
        }
        self.connectivity = {
            "mesh2d": {
                "face_node_connectivity": "mesh2d_face_nodes",
            }
        }
        self.dimensions = {
            "mesh2d": {
                "edge_dimension": "mesh2d_nEdges",
                "face_dimension": "mesh2d_nFaces",
                "node_dimension": "mesh2d_nNodes",
            },
        }

    def test_get_grid_mapping_names(self):
        # Setup doesn't contain any CRS data.
        expected = {"mesh2d": None}
        assert (
            cv._get_grid_mapping_names(self.ds, ["mesh2d"], self.dimensions) == expected
        )
        assert self.ds.ugrid_roles.grid_mapping_names == expected

        # Now add the CRS
        attrs = xugrid.ugrid.crs.crs_to_attrs(pyproj.CRS.from_epsg(28992))
        expected = {"mesh2d": "mesh2d_crs"}
        # Via attrs
        ds = self.ds.copy()
        ds["mesh2d_crs"] = xr.Variable((), 0, attrs=attrs)
        ds["elevation"].attrs["grid_mapping"] = "mesh2d_crs"
        assert cv._get_grid_mapping_names(ds, ["mesh2d"], self.dimensions) == expected
        assert ds.ugrid_roles.grid_mapping_names == expected
        # Via encoding
        ds = self.ds.copy()
        ds["mesh2d_crs"] = xr.Variable((), 0, attrs=attrs)
        ds["elevation"].encoding["grid_mapping"] = "mesh2d_crs"
        assert cv._get_grid_mapping_names(ds, ["mesh2d"], self.dimensions) == expected
        assert ds.ugrid_roles.grid_mapping_names == expected

        # Multiple grid mappings should raise
        ds["elevation2"] = ds["elevation"].copy()
        ds["elevation"].attrs["grid_mapping"] = "mesh2d_crs2"
        with pytest.raises(
            ValueError, match="Multiple grid mappings found for topology"
        ):
            ds.ugrid_roles.grid_mapping_names

    def test_grid_mapping_not_in_dataset(self):
        ds = self.ds.copy()
        ds["elevation"].attrs["grid_mapping"] = "mesh2d_crs"
        expected = {"mesh2d": None}
        with pytest.warns(UserWarning):
            assert (
                cv._get_grid_mapping_names(ds, ["mesh2d"], self.dimensions) == expected
            )

    def test_infer_projected(self):
        ds = self.ds.copy()
        result = cv._infer_projected(ds, ["mesh2d"], self.coordinates)
        assert result == {"mesh2d": True}

        # Test accessor as well
        assert ds.ugrid_roles.is_projected == {"mesh2d": True}

        coordnames = (
            "mesh2d_node_x",
            "mesh2d_node_y",
            "mesh2d_face_x",
            "mesh2d_face_y",
        )
        # Remove two.
        for name in coordnames[:2]:
            ds[name].attrs["standard_name"] = None
        result = cv._infer_projected(ds, ["mesh2d"], self.coordinates)
        assert result == {"mesh2d": True}

        # Remove all.
        for name in coordnames:
            ds[name].attrs["standard_name"] = None
        result = cv._infer_projected(ds, ["mesh2d"], self.coordinates)
        assert result == {"mesh2d": None}

        # Pop two.
        ds = self.ds.copy()
        for name in coordnames[:2]:
            ds[name].attrs.pop("standard_name")
        result = cv._infer_projected(ds, ["mesh2d"], self.coordinates)
        assert result == {"mesh2d": True}

        # Pop all.
        ds = self.ds.copy()
        for name in coordnames:
            ds[name].attrs.pop("standard_name")
        result = cv._infer_projected(ds, ["mesh2d"], self.coordinates)
        assert result == {"mesh2d": None}

        # Make it inconsistent
        ds = self.ds.copy()
        ds["mesh2d_node_x"].attrs["standard_name"] = "longitude"
        ds["mesh2d_node_y"].attrs["standard_name"] = "latitude"
        with pytest.warns(UserWarning):
            cv._infer_projected(ds, ["mesh2d"], self.coordinates)

        # Make it geograhic
        ds["mesh2d_face_x"].attrs["standard_name"] = "longitude"
        ds["mesh2d_face_y"].attrs["standard_name"] = "latitude"
        result = cv._infer_projected(ds, ["mesh2d"], self.coordinates)
        assert result == {"mesh2d": False}


class TestCompleteSpecification:
    """
    This test contains all attributes and variables.

    It is based on some D-Flow output dataset with data on nodes, edges, faces.
    """

    @pytest.fixture(autouse=True)
    def setup(self):
        ds = xr.Dataset()
        ds["mesh2d"] = xr.DataArray(
            0,
            attrs={
                "cf_role": "mesh_topology",
                "long_name": "Topology data of 2D mesh",
                "topology_dimension": 2,
                "node_coordinates": "mesh2d_node_x mesh2d_node_y",
                "node_dimension": "mesh2d_nNodes",
                "max_face_nodes_dimension": "mesh2d_nMax_face_nodes",
                "edge_node_connectivity": "mesh2d_edge_nodes",
                "edge_dimension": "mesh2d_nEdges",
                "edge_coordinates": "mesh2d_edge_x mesh2d_edge_y",
                "face_node_connectivity": "mesh2d_face_nodes",
                "face_dimension": "mesh2d_nFaces",
                "edge_face_connectivity": "mesh2d_edge_faces",
                "face_coordinates": "mesh2d_face_x mesh2d_face_y",
            },
        )
        xy = np.array(
            [
                [0.0, 0.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, 1.0],
            ]
        )
        xy_edge = np.array(
            [
                [0.5, 0.0],
                [1.0, 0.5],
                [0.5, 1.0],
                [0.0, 0.5],
            ]
        )
        x_attrs = {"standard_name": "projection_x_coordinate"}
        y_attrs = {"standard_name": "projection_y_coordinate"}
        ds = ds.assign_coords(
            mesh2d_node_x=xr.DataArray(xy[:, 0], dims=["mesh2d_nNodes"], attrs=x_attrs)
        )
        ds = ds.assign_coords(
            mesh2d_node_y=xr.DataArray(xy[:, 1], dims=["mesh2d_nNodes"], attrs=y_attrs)
        )
        ds = ds.assign_coords(
            mesh2d_edge_x=xr.DataArray(
                xy_edge[:, 0], dims=["mesh2d_nEdges"], attrs=x_attrs
            )
        )
        ds = ds.assign_coords(
            mesh2d_edge_y=xr.DataArray(
                xy_edge[:, 1], dims=["mesh2d_nEdges"], attrs=y_attrs
            )
        )
        ds = ds.assign_coords(
            mesh2d_face_x=xr.DataArray([0.5], dims=["mesh2d_nFaces"], attrs=x_attrs)
        )
        ds = ds.assign_coords(
            mesh2d_face_y=xr.DataArray([0.5], dims=["mesh2d_nFaces"], attrs=y_attrs)
        )
        ds["mesh2d_face_nodes"] = xr.DataArray(
            data=[[0, 1, 2, 3]],
            dims=["mesh2d_nFaces", "mesh2d_nMax_face_nodes"],
            attrs={"_FillValue": -1, "start_index": 0},
        )
        ds["mesh2d_edge_nodes"] = xr.DataArray(
            data=[
                [0, 1],
                [1, 2],
                [2, 3],
                [3, 0],
            ],
            dims=["mesh2d_nEdges", "Two"],
            attrs={"_FillValue": -1, "start_index": 0},
        )
        ds["mesh2d_edge_faces"] = xr.DataArray(
            data=[
                [0, -1],
                [0, -1],
                [0, -1],
                [0, -1],
            ],
            dims=["mesh2d_nEdges", "Two"],
            attrs={"_FillValue": -1, "start_index": 0},
        )
        self.ds = ds

        self.coordinates = {
            "mesh2d": {
                "node_coordinates": (
                    ["mesh2d_node_x"],
                    ["mesh2d_node_y"],
                ),
                "edge_coordinates": (
                    ["mesh2d_edge_x"],
                    ["mesh2d_edge_y"],
                ),
                "face_coordinates": (
                    ["mesh2d_face_x"],
                    ["mesh2d_face_y"],
                ),
            }
        }
        self.connectivity = {
            "mesh2d": {
                "face_node_connectivity": "mesh2d_face_nodes",
                "edge_node_connectivity": "mesh2d_edge_nodes",
                "edge_face_connectivity": "mesh2d_edge_faces",
            }
        }
        self.dimensions = {
            "mesh2d": {
                "edge_dimension": "mesh2d_nEdges",
                "face_dimension": "mesh2d_nFaces",
                "node_dimension": "mesh2d_nNodes",
            },
        }

    def test_topology(self):
        assert self.ds.ugrid_roles.topology == ["mesh2d"]

    def test_coordinates(self):
        assert self.ds.ugrid_roles.coordinates == self.coordinates

    def test_dimensions(self):
        assert self.ds.ugrid_roles.dimensions == self.dimensions

    def test_connectivity(self):
        assert self.ds.ugrid_roles.connectivity == self.connectivity

    def test_dimension_name_mismatch_error(self):
        ds = self.ds.copy()

        ds["mesh2d_edge_nodes"] = xr.DataArray(
            data=[
                [0, 1],
                [1, 2],
                [2, 3],
                [3, 0],
            ],
            dims=["nEdges", "Two"],
            attrs={"_FillValue": -1, "start_index": 0},
        )

        with pytest.raises(
            cv.UgridDimensionError,
            match="edge_dimension: nEdges not in edge_face_connectivity",
        ):
            ds.ugrid_roles.dimensions

    def test_dimension_size_error(self):
        ds = self.ds.copy()

        ds["mesh2d_edge_nodes"] = xr.DataArray(
            data=[
                [0, 1, -1],
                [1, 2, -1],
                [2, 3, -1],
                [3, 0, -1],
            ],
            dims=["mesh2d_nEdges", "Three"],
            attrs={"_FillValue": -1, "start_index": 0},
        )

        with pytest.raises(cv.UgridDimensionError, match="Expected size 2"):
            ds.ugrid_roles.dimensions



"""
This module contains the logic for regridding from a structured form to another
structured form. All coordinates are assumed to be fully orthogonal to each
other.

While the unstructured logic would work for structured data as well, it is much
less efficient than utilizing the structure of the coordinates.
"""

from typing import Any, Tuple, Union

import numpy as np
import xarray as xr

from xugrid.constants import FloatArray, IntArray
from xugrid.regrid.overlap_1d import overlap_1d, overlap_1d_nd
from xugrid.regrid.unstructured import UnstructuredGrid2d
from xugrid.regrid.utils import broadcast
from xugrid.ugrid.ugrid2d import Ugrid2d

# from xugrid import Ugrid2d


class StructuredGrid1d:
    """
    e.g. z -> z; so also works for unstructured

    Parameters
    ----------
    bounds: (n, 2)
    """

    def __init__(self, obj: Union[xr.DataArray, xr.Dataset], name: str):
        bounds_name = f"{name}bounds"  # e.g. xbounds
        size_name = f"d{name}"  # e.g. dx

        index = obj.indexes[name]
        # take care of potentially decreasing coordinate values
        if index.is_monotonic_decreasing:
            midpoints = index.to_numpy()[::-1]
            flipped = True
            side = "right"
        elif index.is_monotonic_increasing:
            midpoints = index.to_numpy()
            flipped = False
            side = "left"
        else:
            raise ValueError(f"{name} is not monotonic for array {obj.name}")

        if bounds_name in obj.coords:
            bounds = obj[bounds_name].to_numpy()
            size = bounds[:, 1] - bounds[:, 0]
        else:
            if size_name in obj.coords:
                # works for scalar size and array size
                size = obj[size_name].to_numpy()
            else:
                # no bounds defined, no dx defined
                # make an estimate of cell size
                size = np.diff(midpoints)
                # Check if equidistant
                atolx = 1.0e-4 * size[0]
                if not np.allclose(size, size[0], atolx):
                    raise ValueError(
                        f"DataArray has to be equidistant along {name}, or "
                        f'explicit bounds must be given as "{name}bounds", or '
                        f'cellsizes must be as "d{name}"'
                    )
                size = np.full_like(midpoints, size[0])

            abs_size = np.abs(size)
            start = midpoints - 0.5 * abs_size
            end = midpoints + 0.5 * abs_size
            bounds = np.column_stack((start, end))

        self.name = name
        self.midpoints = midpoints
        self.bounds = bounds
        self.flipped = flipped
        self.side = side
        self.dname = size_name
        self.dvalue = size
        self.index = index.to_numpy()

    @property
    def coords(self) -> dict:
        coords = {self.name: self.index}
        if self.dvalue.ndim == 0:
            coords[self.dname] = self.dvalue
        else:
            coords[self.dname] = (self.name, self.dvalue)
        return coords

    @property
    def ndim(self) -> int:
        return 1

    @property
    def dims(self) -> Tuple[str]:
        return (self.name,)

    @property
    def size(self) -> int:
        return len(self.bounds)

    @property
    def length(self) -> FloatArray:
        return np.squeeze(abs(np.diff(self.bounds, axis=1)))

    @property
    def directional_bounds(self):
        # Only flip bounds if needed
        if self.flipped:
            return self.bounds[::-1, :].copy()
        else:
            return self.bounds

    def flip_if_needed(self, index: IntArray) -> IntArray:
        if self.flipped:
            return self.size - index - 1
        else:
            return index

    def valid_nodes_within_bounds(
        self, other: "StructuredGrid1d"
    ) -> Tuple[IntArray, IntArray]:
        """
        Return nodes when midpoints are within bounding box of overlaying grid.

        In cases that midpoints (and bounding boxes) are flipped, computed indexes
        are fliped as well.

        Parameters
        ----------
        other: StructuredGrid1d
            The target grid topology

        Returns
        -------
        valid_self_index: np.array
            valid source indexes
        valid_other_index: np.array
            valid target indexes
        """
        start = np.searchsorted(self.bounds[:, 0], other.midpoints, side=self.side)
        end = np.searchsorted(self.bounds[:, 1], other.midpoints, side=self.side)
        valid = (
            (start == (end + 1))
            & (other.midpoints > self.bounds[0, 0])
            & (other.midpoints < self.bounds[-1, 1])
        )
        valid_other_index = np.arange(other.size)[valid]
        valid_self_index = end[valid]
        valid_self_index = self.flip_if_needed(valid_self_index)
        valid_other_index = other.flip_if_needed(valid_other_index)
        return valid_self_index, valid_other_index

    def valid_nodes_within_bounds_and_extend(
        self, other: "StructuredGrid1d"
    ) -> Tuple[IntArray, IntArray]:
        """
        Return all valid nodes for linear interpolation.
        In addition to valid_nodes_within_bounds() is checked if target
        midpoints are not outside outer source boundary midpoints. In that case
        there is no interpolation possible.

        Parameters
        ----------
        other: StructuredGrid1d
            The target grid.

        Returns
        -------
        valid_self_index: np.array
            valid source indexes
        valid_other_index: np.array
            valid target indexes
        """
        source_index, target_index = self.valid_nodes_within_bounds(other)
        return source_index, target_index
        valid = (other.midpoints[target_index] > self.midpoints[0]) & (
            other.midpoints[target_index] < self.midpoints[-1]
        )
        return source_index[valid], target_index[valid]

    def overlap_1d_structured(
        self, other: "StructuredGrid1d"
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Return source and target nodes and overlapping length. It utilises overlap_1d()
        and does an aditional flip in cases of reversed midpoints

        Parameters
        ----------
        other: StructuredGrid1d
            The target grid.

        Returns
        -------
        valid_self_index: np.array
            valid source indexes
        valid_other_index: np.array
            valid target indexes
        weights: np.array
            length of overlap
        """
        source_index, target_index, weights = overlap_1d(self.bounds, other.bounds)
        source_index = self.flip_if_needed(source_index)
        target_index = other.flip_if_needed(target_index)
        return source_index, target_index, weights

    def centroids_to_linear_sets(
        self,
        source_index: np.array,
        target_index: np.array,
        weights: np.array,
        neighbour: np.array,
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Return for every target node an pair of connected source nodes based
        on centroids connection inputs.

        Parameters
        ----------
        source_index: np.array
        target_index: np.array
        weights: np.array

        Returns
        -------
        valid_self_index: np.array
        valid_other_index: np.array
        weights: np.array
            lineair interpolation weights.
        """
        # if source_index is flipped, source-index is decreasing and neighbour
        # need to be flipped
        if self.flipped:
            neighbour = -neighbour
        source_index = np.column_stack((source_index, source_index + neighbour)).ravel()
        target_index = np.repeat(target_index, 2)
        weights = np.column_stack((weights, 1.0 - weights)).ravel()

        # correct for possibility of out of bound due to column-stack
        # source_index + 1 and -1
        valid = np.logical_and(source_index <= self.size - 1, source_index >= 0)
        return source_index[valid], target_index[valid], weights[valid]

    def maybe_reverse_index(self, index: IntArray) -> IntArray:
        """
        Flips the index if needed for descending coordinates.

        Parameters
        ----------
        index: np.ndarray

        Returns
        -------
        checked_index: np.ndarray
        """
        if self.flipped:
            return self.size - index - 1
        else:
            return index

    def compute_linear_weights_to_centroids(
        self, other: "StructuredGrid1d", source_index: IntArray, target_index: IntArray
    ) -> Tuple[FloatArray, IntArray]:
        """
        Compute linear weights bases on centroid indexes.

        Parameters
        ----------
        other: StructuredGrid1d
        source_index: np.array
        target_index: np.array

        Raises
        ------
        ValueError
            When the coordinate contains only a single point.

        Returns
        -------
        weights: np.array
        neighbor: np.narray
        """
        if self.midpoints.size < 2:
            raise ValueError(
                f"Coordinate {self.name} has size: {self.midpoints.size}. "
                "At least two points are required for interpolation."
            )

        source_midpoint_index = self.maybe_reverse_index(source_index)
        target_midpoint_index = other.maybe_reverse_index(target_index)
        # cases where midpoint target <= midpoint source: set neighbor to -1
        neighbor = np.where(
            other.midpoints[target_midpoint_index]
            <= self.midpoints[source_midpoint_index],
            -1,
            1,
        )

        # When we exceed the original domain, it should still interpolate
        # within the bounds.
        # Make sure neighbor falls in [0, n)
        neighbor_index = np.clip(
            source_midpoint_index + neighbor, 0, self.midpoints.size - 1
        )
        # Update neighbor since we return it
        neighbor = neighbor_index - source_midpoint_index

        # If neighbor is 0, we end up computing zero distance, since we're
        # comparing a midpoint to iself. Instead, set a weight of 1.0 on one,
        # (and impliclity 0 in the other). Similarly, if source and target
        # midpoints coincide, the distance may end up 0.
        length = (
            other.midpoints[target_midpoint_index]
            - self.midpoints[source_midpoint_index]
        )
        total_length = (
            self.midpoints[neighbor_index] - self.midpoints[source_midpoint_index]
        )
        # Do not divide by zero.
        # We will overwrite the value anyway at neighbor == 0.
        total_length[total_length == 0] = 1
        weights = 1 - (length / total_length)
        weights[neighbor == 0] = 0.0
        condition = np.logical_and(weights < 0.0, weights > 1.0)
        if condition.any():
            raise ValueError(
                f"Computed invalid weights for dimensions: {self.name} at coords: {self.midpoints[condition]}"
            )
        return weights, neighbor

    def sorted_output(
        self, source_index: IntArray, target_index: IntArray, weights: FloatArray
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Return sorted input based on target index. The regridder needs input
        sorted on row index of WeightMatrixCOO (target index).

        Parameters
        ----------
        source_index: np.array
        target_index: np.array
        weights: np.array

        Returns
        -------
        source_index: np.array
        target_index: np.array
        weights: np.array
        """
        sorter_target = np.argsort(target_index)
        return (
            source_index[sorter_target],
            target_index[sorter_target],
            weights[sorter_target],
        )

    def overlap(
        self, other: "StructuredGrid1d", relative: bool
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Return source and target indexes and overlapping length

        Parameters
        ----------
        other: StructuredGrid1d
        relative: bool
            Overlapping length target relative to source length

        Returns
        -------
        source_index: np.array
        target_index: np.array
        weights: np.array
            Overlapping length
        """
        source_index, target_index, weights = self.overlap_1d_structured(other)
        if relative:
            weights /= self.length[source_index]
        return self.sorted_output(source_index, target_index, weights)

    def locate_centroids(
        self,
        other: "StructuredGrid1d",
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Return source and target indexes based on nearest neighbor of
        centroids.

        Parameters
        ----------
        other: StructuredGrid1d

        Returns
        -------
        source_index: np.array
        target_index: np.array
        weights: np.array
        """
        source_index, target_index = self.valid_nodes_within_bounds(other)
        weights = np.ones(source_index.size, dtype=float)
        return self.sorted_output(source_index, target_index, weights)

    def linear_weights(
        self, other: "StructuredGrid1d"
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Return linear source and target indexes and corresponding weights.

        Parameters
        ----------
        other: StructuredGrid1d

        Returns
        -------
        source_index: np.array
        target_index: np.array
        weights: np.array
        """
        source_index, target_index = self.valid_nodes_within_bounds_and_extend(other)
        weights, neighbour = self.compute_linear_weights_to_centroids(
            other, source_index, target_index
        )
        source_index, target_index, weights = self.centroids_to_linear_sets(
            source_index,
            target_index,
            weights,
            neighbour,
        )
        return self.sorted_output(source_index, target_index, weights)

    def to_dataset(self, name: str) -> xr.DataArray:
        export_name = name + "_" + self.name
        return xr.DataArray(
            name=name,
            data=np.nan,
            dims=[export_name, export_name + "nbounds"],
            coords={
                export_name: self.midpoints,
                export_name + "bounds": (
                    [export_name, export_name + "nbounds"],
                    self.bounds,
                ),
                export_name + "nbounds": np.arange(2),
            },
        )


class StructuredGrid2d(StructuredGrid1d):
    """Represent e.g. raster data topology."""

    def __init__(
        self,
        obj: Union[xr.DataArray, xr.Dataset],
        name_x: str,
        name_y: str,
    ):
        self.xbounds = StructuredGrid1d(obj, name_x)
        self.ybounds = StructuredGrid1d(obj, name_y)

    @property
    def coords(self) -> dict:
        return self.ybounds.coords | self.xbounds.coords

    @property
    def ndim(self) -> int:
        return 2

    @property
    def dims(self) -> Tuple[str, str]:
        return self.ybounds.dims + self.xbounds.dims  # ("y", "x")

    @property
    def size(self) -> int:
        return self.ybounds.size * self.xbounds.size

    @property
    def shape(self) -> Tuple[int, int]:
        return (self.ybounds.size, self.xbounds.size)

    @property
    def area(self) -> FloatArray:
        return np.multiply.outer(self.ybounds.length, self.xbounds.length)

    def convert_to(self, matched_type: Any) -> Any:
        if matched_type == StructuredGrid2d:
            return self
        elif matched_type == UnstructuredGrid2d:
            ugrid2d = Ugrid2d.from_structured_bounds(
                self.xbounds.directional_bounds,
                self.ybounds.directional_bounds,
            )
            return UnstructuredGrid2d(ugrid2d)
        else:
            raise TypeError(
                f"Cannot convert StructuredGrid2d to {matched_type.__name__}"
            )

    def broadcast_sorted(
        self,
        other: Any,
        source_index_y: IntArray,
        source_index_x: IntArray,
        target_index_y: IntArray,
        target_index_x: IntArray,
        weights_y: FloatArray,
        weights_x: FloatArray,
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        source_index, target_index, weights = broadcast(
            self.shape,
            other.shape,
            (source_index_y, source_index_x),
            (target_index_y, target_index_x),
            (weights_y, weights_x),
        )
        sorter = np.argsort(target_index)
        return source_index[sorter], target_index[sorter], weights[sorter]

    def overlap(self, other, relative: bool) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Compute (relative) overlap with other.

        Returns
        -------
        source_index: 1d np.ndarray of int
        target_index: 1d np.ndarray of int
        weights: 1d np.ndarray of float
        """
        source_index_x, target_index_x, weights_x = self.xbounds.overlap(
            other.xbounds, relative
        )
        source_index_y, target_index_y, weights_y = self.ybounds.overlap(
            other.ybounds, relative
        )
        return self.broadcast_sorted(
            other,
            source_index_y,
            source_index_x,
            target_index_y,
            target_index_x,
            weights_y,
            weights_x,
        )

    def locate_centroids(
        self, other, tolerance
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Locate centroids of other in self.

        Returns
        -------
        source_index: 1d np.ndarray of int
        target_index: 1d np.ndarray of int
        weights: 1d np.ndarray of float
        """
        source_index_x, target_index_x, weights_x = self.xbounds.locate_centroids(
            other.xbounds
        )
        source_index_y, target_index_y, weights_y = self.ybounds.locate_centroids(
            other.ybounds
        )
        return self.broadcast_sorted(
            other,
            source_index_y,
            source_index_x,
            target_index_y,
            target_index_x,
            weights_y,
            weights_x,
        )

    def linear_weights(self, other) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Compute linear interpolation weights with other.

        Returns
        -------
        source_index: 1d np.ndarray of int
        target_index: 1d np.ndarray of int
        weights: 1d np.ndarray of float
        """
        source_index_x, target_index_x, weights_x = self.xbounds.linear_weights(
            other.xbounds
        )
        source_index_y, target_index_y, weights_y = self.ybounds.linear_weights(
            other.ybounds
        )
        return self.broadcast_sorted(
            other,
            source_index_y,
            source_index_x,
            target_index_y,
            target_index_x,
            weights_y,
            weights_x,
        )

    def to_dataset(self, name: str) -> xr.Dataset:
        ds_x = self.xbounds.to_dataset(name)
        ds_y = self.ybounds.to_dataset(name)
        ds = xr.merge([ds_x, ds_y])
        ds[name + "_type"] = xr.DataArray(-1, attrs={"type": "StructuredGrid2d"})
        return ds


class StructuredGrid3d:
    """
    e.g. (x,y,z) -> (x,y,z)

    A voxel model (GeoTOP)
    """

    def __init__(
        self,
        obj: Union[xr.DataArray, xr.Dataset],
        name_x: str,
        name_y: str,
        name_z: str,
    ):
        self.xbounds = StructuredGrid1d(obj, name_x)
        self.ybounds = StructuredGrid1d(obj, name_y)
        self.zbounds = StructuredGrid1d(obj, name_z)

    @property
    def shape(self):
        return (self.zbound.size, self.ybounds.size, self.xbounds.size)

    @property
    def volume(self):
        return np.multiply.outer(self.zbounds.length, self.area)

    def broadcast_sorted(
        self,
        other,
        source_index_z: np.array,
        source_index_y: np.array,
        source_index_x: np.array,
        target_index_z: np.array,
        target_index_y: np.array,
        target_index_x: np.array,
        weights_z: np.array,
        weights_y: np.array,
        weights_x: np.array,
    ):
        source_index, target_index, weights = broadcast(
            self.shape,
            other.shape,
            (source_index_z, source_index_y, source_index_x),
            (target_index_z, target_index_y, target_index_x),
            (weights_z, weights_y, weights_x),
        )
        sorter = np.argsort(target_index)
        return source_index[sorter], target_index[sorter], weights[sorter]

    def overlap(self, other, relative: bool):
        """
        Compute (relative) overlap with other.

        Returns
        -------
        source_index: 1d np.ndarray of int
        target_index: 1d np.ndarray of int
        weights: 1d np.ndarray of float
        """
        source_index_x, target_index_x, weights_x = self.xbounds.overlap(
            other.xbounds, relative
        )
        source_index_y, target_index_y, weights_y = self.ybounds.overlap(
            other.ybounds, relative
        )
        source_index_z, target_index_z, weights_z = self.zbounds.overlap(
            other.zbounds, relative
        )
        return self.broadcast_sorted(
            other,
            source_index_z,
            source_index_y,
            source_index_x,
            target_index_z,
            target_index_y,
            target_index_x,
            weights_z,
            weights_y,
            weights_x,
        )

    def locate_centroids(self, other):
        source_index_x, target_index_x, weights_x = self.xbounds.locate_centroids(
            other.xbounds
        )
        source_index_y, target_index_y, weights_y = self.ybounds.locate_centroids(
            other.ybounds
        )
        source_index_z, target_index_z, weights_z = self.zbounds.locate_centroids(
            other.zbounds
        )
        return self.broadcast_sorted(
            other,
            source_index_z,
            source_index_y,
            source_index_x,
            target_index_z,
            target_index_y,
            target_index_x,
            weights_z,
            weights_y,
            weights_x,
        )

    def linear_weights(self, other):
        source_index_x, target_index_x, weights_x = self.xbounds.linear_weights(
            other.xbounds
        )
        source_index_y, target_index_y, weights_y = self.ybounds.linear_weights(
            other.ybounds
        )
        source_index_z, target_index_z, weights_z = self.zbounds.linear_weights(
            other.zbounds
        )
        return self.broadcast_sorted(
            other,
            source_index_z,
            source_index_y,
            source_index_x,
            target_index_z,
            target_index_y,
            target_index_x,
            weights_z,
            weights_y,
            weights_x,
        )


class ExplicitStructuredGrid3d:
    """
    e.g. (x,y,z) -> (x,y,z)

    A layered model (E.g. REGIS)
    """

    def __init__(
        self,
        obj: Union[xr.DataArray, xr.Dataset],
    ):
        # zbounds is a 3D array with dimensions (nlayer, y.size * x.size, 2)
        self.xbounds = StructuredGrid1d(obj, "x")
        self.ybounds = StructuredGrid1d(obj, "y")
        self.zbounds = StructuredGrid1d(obj, "z")

    @property
    def shape(self):
        raise NotImplementedError

    @property
    def volume(self):
        return np.multiply.outer(self.zbounds.length, self.area)

    def overlap(self, other, relative: bool):
        """
        Compute (relative) overlap with other.

        Returns
        -------
        source_index: 1d np.ndarray of int
        target_index: 1d np.ndarray of int
        weights: 1d np.ndarray of float
        """
        source_index_x, target_index_x, weights_x = self.xbounds.overlap(
            other.xbounds, relative
        )
        source_index_y, target_index_y, weights_y = self.ybounds.overlap(
            other.ybounds, relative
        )
        source_index_yx, target_index_yx, weights_yx = broadcast(
            self.shape,
            other.shape,
            (source_index_y, source_index_x),
            (target_index_y, target_index_x),
            (weights_y, weights_x),
        )

        if isinstance(other, StructuredGrid3d):
            zbounds = other.zbounds[np.newaxis, ...]
            target_index = np.zeros(zbounds.shape[1], dtype=int)
        elif isinstance(other, ExplicitStructuredGrid3d):
            zbounds = other.zbounds
            target_index = target_index_yx
        else:
            raise TypeError

        source_index_zyx, target_index_zyx, weights_z = overlap_1d_nd(
            self.zbounds,
            zbounds,
            source_index_yx,
            target_index,
        )
        # TODO: check array dims
        weights_zyx = weights_z * weights_yx
        return source_index_zyx, target_index_zyx, weights_zyx



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pyproj


class CrsPlaceholder:
    """Stands in for pyproj.CRS when pyproj is not installed."""

    def __init__(self, attrs: dict):
        self._attrs = dict(attrs)

    def __eq__(self, other):
        if isinstance(other, CrsPlaceholder):
            return self._attrs == other._attrs
        return False

    def __repr__(self):
        return f"CrsPlaceholder({self._attrs})"


def crs_from_attrs(ds_attrs: dict) -> pyproj.CRS | CrsPlaceholder:
    """
    Attempt to construct a pyproj.CRS from grid mapping attributes.

    Tries in order:
    1. CF grid mapping attributes
    2. WKT (crs_wkt or spatial_ref)
    3. EPSG code

    Returns CrsPlaceholder if no valid CRS can be determined.
    Note that it looks like the from_cf method prefers the WKT
    form even if the name is given.
    """
    try:
        import pyproj
    except ImportError:
        return CrsPlaceholder(ds_attrs)

    # 0. Lower keys just in case.
    attrs = {k.lower(): v for k, v in ds_attrs.items()}

    # 1. CF grid mapping attributes
    name = attrs.get("grid_mapping_name")
    if name is not None:
        try:
            return pyproj.CRS.from_cf(attrs)
        except pyproj.exceptions.CRSError:
            pass

    # 2. WKT
    wkt = attrs.get("crs_wkt") or attrs.get("spatial_ref")
    if wkt is not None:
        try:
            return pyproj.CRS.from_wkt(wkt)
        except pyproj.exceptions.CRSError:
            pass

    # 3. EPSG fallback
    epsg_entry = attrs.get("epsg") or attrs.get("epsg_code")
    if epsg_entry is not None:
        try:
            return pyproj.CRS.from_user_input(epsg_entry)
        except (ValueError, pyproj.exceptions.CRSError):
            pass

    # 4. pyproj couldn't make sense of it. Return a placeholder instead.
    return CrsPlaceholder(ds_attrs)


def crs_to_attrs(crs: pyproj.CRS | CrsPlaceholder) -> dict:
    if isinstance(crs, CrsPlaceholder):
        return crs._attrs

    attrs = crs.to_cf()  # already includes crs_wkt and CF attrs when possible.
    # GDAL compat: GDAL uses spatial_ref also.
    attrs["spatial_ref"] = attrs["crs_wkt"]
    attrs["name"] = crs.name
    epsg = crs.to_epsg()
    if epsg is not None:
        attrs["epsg"] = epsg
    return attrs



import abc
from typing import Tuple, Union

import numpy as np
import xarray as xr

import xugrid


class AbstractUgridAccessor(abc.ABC):
    @abc.abstractmethod
    def to_dataset(self):
        pass

    @abc.abstractmethod
    def assign_node_coords(self):
        pass

    @abc.abstractmethod
    def set_node_coords(self):
        pass

    @property
    @abc.abstractmethod
    def crs(self):
        pass

    @abc.abstractmethod
    def set_crs(self):
        pass

    @abc.abstractmethod
    def to_crs(self):
        pass

    @abc.abstractmethod
    def sel(self):
        pass

    @abc.abstractmethod
    def sel_points(self, x, y, method, out_of_bounds, fill_value, tolerance):
        pass

    @abc.abstractmethod
    def intersect_line(self):
        pass

    @abc.abstractmethod
    def intersect_linestring(self):
        pass

    @property
    @abc.abstractmethod
    def bounds(self):
        pass

    @property
    @abc.abstractmethod
    def total_bounds(self):
        pass

    @property
    @abc.abstractmethod
    def name(self):
        pass

    @property
    @abc.abstractmethod
    def names(self):
        pass

    @property
    @abc.abstractmethod
    def topology(self):
        pass

    @staticmethod
    def _raster_xy(bounds: Tuple[float, float, float, float], resolution: float):
        xmin, ymin, xmax, ymax = bounds
        d = abs(resolution)
        xmin = np.floor(xmin / d) * d
        xmax = np.ceil(xmax / d) * d
        ymin = np.floor(ymin / d) * d
        ymax = np.ceil(ymax / d) * d
        x = np.arange(xmin + 0.5 * d, xmax, d)
        y = np.arange(ymax - 0.5 * d, ymin, -d)
        return x, y

    def _raster(self, x, y, index) -> xr.DataArray:
        index = index.ravel()
        indexer = xr.DataArray(
            data=index.reshape(y.size, x.size),
            coords={"y": y, "x": x},
            dims=["y", "x"],
        )
        out = self.obj.isel({self.grid.face_dimension: indexer}).where(indexer != -1)
        return out

    def clip_box(
        self,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
    ):
        """
        Clip the DataArray or Dataset by a bounding box.

        Parameters
        ----------
        xmin: float
        ymin: float
        xmax: float
        ymax: float

        -------
        clipped:
            xugrid.UgridDataArray or xugrid.UgridDataset
        """
        return self.sel(x=slice(xmin, xmax), y=slice(ymin, ymax))

    def partition_by_label(
        self, labels: np.ndarray
    ) -> Union["xugrid.UgridDataArray", "xugrid.UgridDataset"]:
        """
        Partition a grid by labels.

        Parameters
        ----------
        labels: np.ndarray of integers labeling each face.

        Returns
        -------
        partitioned: list of partitions
        """
        from xugrid.ugrid import partitioning

        return partitioning.partition_by_label(self.grid, self.obj, labels)

    def partition(
        self, n_part: int
    ) -> Union["xugrid.UgridDataArray", "xugrid.UgridDataset"]:
        """
        Partition a grid into a given number of parts.

        Parameters
        ----------
        n_part: integer
            The number of parts to partition the mesh.

        Returns
        -------
        partitioned: list of partitions
        """
        labels = self.grid.label_partitions(n_part)
        return self.partition_by_label(labels)

    def to_netcdf(self, *args, **kwargs):
        """
        Write dataset contents to a UGRID compliant netCDF file.

        This function wraps :py:meth:`xr.Dataset.to_netcdf`; it adds the UGRID
        variables and coordinates to a standard xarray Dataset, then writes the
        result to a netCDF.

        All arguments are forwarded to :py:meth:`xr.Dataset.to_netcdf`.
        """
        self.to_dataset().to_netcdf(*args, **kwargs)

    def to_zarr(self, *args, **kwargs):
        """
        Write dataset contents to a UGRID compliant Zarr file.

        This function wraps :py:meth:`xr.Dataset.to_zarr`; it adds the UGRID
        variables and coordinates to a standard xarray Dataset, then writes the
        result to a Zarr file.

        All arguments are forwarded to :py:meth:`xr.Dataset.to_zarr`.
        """
        self.to_dataset().to_zarr(*args, **kwargs)



"""Contains common reduction methods."""

import math
from typing import Callable

import numba as nb
import numpy as np

from xugrid.regrid.nanpercentile import _select_two


def mean(values, weights, workspace):
    vsum = 0.0
    wsum = 0.0
    for v, w in zip(values, weights):
        if np.isnan(v):
            continue
        vsum += w * v
        wsum += w
    if wsum == 0:
        return np.nan
    else:
        return vsum / wsum


def harmonic_mean(values, weights, workspace):
    v_agg = 0.0
    w_sum = 0.0
    for v, w in zip(values, weights):
        if np.isnan(v) or v == 0:
            continue
        if w > 0:
            w_sum += w
            v_agg += w / v
    if v_agg == 0 or w_sum == 0:
        return np.nan
    else:
        return w_sum / v_agg


def geometric_mean(values, weights, workspace):
    v_agg = 0.0
    w_sum = 0.0

    # Compute sum to normalize weights to avoid tiny or huge values in exp
    normsum = 0.0
    for v, w in zip(values, weights):
        normsum += w
    # Early return if no values
    if normsum == 0:
        return np.nan

    for v, w in zip(values, weights):
        w = w / normsum
        # Skip if v is NaN or 0.
        if v > 0 and w > 0:
            v_agg += w * np.log(abs(v))
            w_sum += w
        elif v < 0:
            # Computing a geometric mean of negative numbers requires a complex
            # value.
            return np.nan

    if w_sum == 0:
        return np.nan
    else:
        # w_sum is generally 1.0, but might not be if there are NaNs present!
        return np.exp((1.0 / w_sum) * v_agg)


def sum(values, weights, workspace):
    v_sum = 0.0
    w_sum = 0.0

    for v, w in zip(values, weights):
        if np.isnan(v):
            continue
        v_sum += v
        w_sum += w
    if w_sum == 0:
        return np.nan
    else:
        return v_sum


@nb.njit(inline="always")
def _minimum(values, weights):
    v_min = np.inf
    w_max = 0.0
    for v, w in zip(values, weights):
        if np.isnan(v):
            continue
        v_min = min(v, v_min)
        w_max = max(w_max, w)
    if w_max == 0.0:
        return np.nan
    return v_min


def minimum(values, weights, workspace):
    return _minimum(values, weights)


@nb.njit(inline="always")
def _maximum(values, weights):
    v_max = -np.inf
    w_max = 0.0
    for v, w in zip(values, weights):
        if np.isnan(v):
            continue
        v_max = max(v, v_max)
        w_max = max(w_max, w)
    if w_max == 0.0:
        return np.nan
    return v_max


def maximum(values, weights, workspace):
    return _maximum(values, weights)


def mode(values, weights, workspace):
    # Area weighted mode. We use a linear search to accumulate weights, as we
    # generally expect a relatively small number of elements in the indices and
    # weights arrays.
    accum = workspace
    accum[: weights.size] = weights[:]
    w_sum = 0
    w_max = 0.0
    for running_total, (v, w) in enumerate(zip(values, weights)):
        if np.isnan(v):
            continue
        w_max = max(w, w_max)
        w_sum += 1
        for j in range(running_total):  # Compare with previously found values
            if values[j] == v:  # matches previous value
                accum[j] += w  # increase previous weight sum
                break

    if w_sum == 0 or w_max == 0.0:
        # Everything skipped (all nodata), or all weights zero
        return np.nan
    else:
        # Find value with highest frequency.
        # In case frequencies are equal (a tie), take the largest value.
        # This ensures the same result irrespective of value ordering.
        w_max = 0
        mode_value = values[0]
        for w_accum, v in zip(accum, values):
            if ~np.isnan(v):
                if (w_accum > w_max) or (w_accum == w_max and v > mode_value):
                    w_max = w_accum
                    mode_value = v
        return mode_value


@nb.njit
def percentile(values, weights, workspace, p):
    # This function is a simplified port of:
    # https://github.com/numba/numba/blob/0441bb17c7820efc2eba4fd141b68dac2afa4740/numba/np/arraymath.py#L1745

    # Exit early if all weights are 0.
    w_max = 0.0
    for w in weights:
        w_max = max(w, w_max)
    if w_max == 0.0:
        return np.nan

    if p == 0:
        return _minimum(values, weights)

    if p == 100:
        return _maximum(values, weights)

    # Everything should've been checked before:
    #
    # * a.dtype should be float
    # * 0 <= q <= 100.
    #
    # Filter the NaNs

    n = 0
    for v in values:
        if ~np.isnan(v):
            workspace[n] = v
            n += 1

    # Early returns
    if n == 0:
        return np.nan
    if n == 1:
        return workspace[0]

    # linear interp between closest ranks
    rank = 1 + (n - 1) * p / 100.0
    f = math.floor(rank)
    m = rank - f
    lower, upper = _select_two(workspace[:n], k=int(f - 1), low=0, high=(n - 1))
    return lower * (1 - m) + upper * m


def first_order_conservative(values, weights, workspace):
    # Uses relative weights!
    # Rename to: first order conservative?
    v_agg = 0.0
    w_sum = 0.0
    for v, w in zip(values, weights):
        if np.isnan(v):
            continue
        v_agg += v * w
        w_sum += w
    if w_sum == 0:
        return np.nan
    else:
        return v_agg


conductance = first_order_conservative


def max_overlap(values, weights, workspace):
    w_max = 0.0
    v_max = -np.inf
    # Find value with highest overlap.
    # In case frequencies are equal (a tie), take the largest value.
    # This ensures the same result irrespective of value ordering.
    for v, w in zip(values, weights):
        if ~np.isnan(v):
            if (w > w_max) or (w == w_max and v > v_max):
                w_max = w
                v_max = v
    if w_max == 0.0:
        return np.nan
    return v_max


def create_percentile_method(p: float) -> Callable:
    if not (0.0 <= p <= 100.0):
        raise ValueError(f"percentile must be in the range [0, 100], received: {p}")

    def f(values, weights, workspace) -> float:
        return percentile(values, weights, workspace, p)

    return f


median = create_percentile_method(50)


ABSOLUTE_OVERLAP_METHODS = {
    "mean": mean,
    "harmonic_mean": harmonic_mean,
    "geometric_mean": geometric_mean,
    "sum": sum,
    "minimum": minimum,
    "maximum": maximum,
    "mode": mode,
    "median": median,
    "max_overlap": max_overlap,
}
for p in (5, 10, 25, 50, 75, 90, 95):
    ABSOLUTE_OVERLAP_METHODS[f"p{p}"] = create_percentile_method(p)


RELATIVE_OVERLAP_METHODS = {
    "conductance": conductance,
    "first_order_conservative": first_order_conservative,
}



import numpy as np
import pytest
import xarray as xr

import xugrid as xu


@pytest.fixture(scope="function")
def structured_grid():
    dims = ("y", "x")
    y = np.arange(3.5, -0.5, -1.0)
    x = np.arange(0.5, 4.5, 1.0)
    coords = {"y": y, "x": x}

    return xr.DataArray(np.ones((4, 4), dtype=np.int32), coords=coords, dims=dims)


@pytest.fixture(scope="function")
def unstructured_grid(structured_grid):
    return xu.UgridDataArray.from_structured2d(structured_grid)


@pytest.fixture(scope="function")
def network():
    node_xy = np.array(
        [
            [0.0, 0.0],
            [1.5, 1.5],
            [2.5, 1.5],
            [4.0, 0.0],
            [4.0, 3.0],
        ]
    )
    edge_nodes = np.array(
        [
            [0, 1],
            [1, 2],
            [2, 3],
            [2, 4],
        ]
    )
    ugrid1d = xu.Ugrid1d(*node_xy.T, -1, edge_nodes)
    data = xr.DataArray(
        np.array([1, 2, 4, -4], dtype=float), dims=(ugrid1d.edge_dimension,)
    )
    return xu.UgridDataArray(data, grid=ugrid1d)


@pytest.fixture(scope="function")
def points_to_sample():
    x_loc = np.array([0.5, 1.5, 2.5, 3.5, 3.5])
    y_loc = np.array([0.5, 1.5, 1.5, 2.5, 0.5])
    diag = 0.5 * np.sqrt(2)
    expected_values = np.array(
        [
            1.0,
            (diag * 1 + 0.5 * 2) / (diag + 0.5),  # 1 diagonal edge, 1 horizontal edge
            (0.5 * 2 + diag * -4 + diag * 4)
            / (2 * diag + 0.5),  # 2 diagonal edge, 1 horizontal edge
            -4.0,
            4.0,
        ]
    )

    return x_loc, y_loc, expected_values


def test_network_gridder_init__unstructured(network, unstructured_grid):
    gridder = xu.NetworkGridder(network, unstructured_grid, method="mean")

    assert isinstance(gridder, xu.NetworkGridder)
    assert gridder._source.ugrid_topology == network.grid
    assert gridder._target.ugrid_topology == unstructured_grid.grid
    assert gridder._weights.n == unstructured_grid.grid.n_face
    assert gridder._weights.m == network.grid.n_edge
    assert gridder._weights.nnz == 8


def test_network_gridder_regrid__unstructured(
    network, unstructured_grid, points_to_sample
):
    gridder = xu.NetworkGridder(network, unstructured_grid, method="mean")
    gridded = gridder.regrid(network)

    assert isinstance(gridded, type(unstructured_grid))
    assert gridded.shape == unstructured_grid.shape
    assert np.count_nonzero(np.isnan(gridded)) == 11

    x_loc, y_loc, expected_values = points_to_sample
    grid_values = gridded.ugrid.sel_points(x=x_loc, y=y_loc)

    np.testing.assert_allclose(grid_values, expected_values)


def test_network_gridder_regrid__unstructured_transient(
    network, unstructured_grid, points_to_sample
):
    # Make transient network
    times = [np.datetime64("2022-01-01"), np.datetime64("2023-01-01")]
    time_multiplier = xr.DataArray([1.0, 2.0], dims="time", coords={"time": times})
    network = (network * time_multiplier).transpose(
        "time", network.ugrid.grid.core_dimension
    )

    gridder = xu.NetworkGridder(network, unstructured_grid, method="mean")
    gridded = gridder.regrid(network)

    assert isinstance(gridded, type(unstructured_grid))
    assert np.count_nonzero(np.isnan(gridded)) == 22

    x_loc, y_loc, expected_values = points_to_sample
    grid_values_t0 = gridded.isel(time=0).ugrid.sel_points(x=x_loc, y=y_loc)
    grid_values_t1 = gridded.isel(time=1).ugrid.sel_points(x=x_loc, y=y_loc)

    np.testing.assert_allclose(grid_values_t0, expected_values)
    np.testing.assert_allclose(grid_values_t1, 2 * expected_values)


def test_network_gridder_init__structured(network, structured_grid):
    gridder = xu.NetworkGridder(network, structured_grid, method="mean")

    assert isinstance(gridder, xu.NetworkGridder)
    assert gridder._source.ugrid_topology == network.grid
    np.testing.assert_array_equal(
        gridder._target.coords["x"], structured_grid.x.to_numpy()
    )
    np.testing.assert_array_equal(
        gridder._target.coords["y"], structured_grid.y.to_numpy()
    )
    assert gridder._weights.n == structured_grid.size
    assert gridder._weights.m == network.grid.n_edge
    assert gridder._weights.nnz == 8


def test_network_gridder_regrid__structured(network, structured_grid, points_to_sample):
    gridder = xu.NetworkGridder(network, structured_grid, method="mean")
    gridded = gridder.regrid(network)

    assert isinstance(gridded, type(structured_grid))
    assert gridded.shape == structured_grid.shape
    assert np.count_nonzero(np.isnan(gridded)) == 11

    x_loc, y_loc, expected_values = points_to_sample
    x_loc = xr.DataArray(x_loc, dims="points")
    y_loc = xr.DataArray(y_loc, dims="points")
    grid_values = gridded.sel(x=x_loc, y=y_loc).to_numpy()

    np.testing.assert_allclose(grid_values, expected_values)


def test_network_gridder_regrid__structured_transient(
    network, structured_grid, points_to_sample
):
    # Make transient network
    times = [np.datetime64("2022-01-01"), np.datetime64("2023-01-01")]
    time_multiplier = xr.DataArray([1.0, 2.0], dims="time", coords={"time": times})
    network = (network * time_multiplier).transpose(
        "time", network.ugrid.grid.core_dimension
    )

    gridder = xu.NetworkGridder(network, structured_grid, method="mean")
    gridded = gridder.regrid(network)

    assert isinstance(gridded, type(structured_grid))
    assert np.count_nonzero(np.isnan(gridded)) == 22

    x_loc, y_loc, expected_values = points_to_sample
    x_loc = xr.DataArray(x_loc, dims="points")
    y_loc = xr.DataArray(y_loc, dims="points")
    grid_values_t0 = gridded.isel(time=0).sel(x=x_loc, y=y_loc).to_numpy()
    grid_values_t1 = gridded.isel(time=1).sel(x=x_loc, y=y_loc).to_numpy()

    np.testing.assert_allclose(grid_values_t0, expected_values)
    np.testing.assert_allclose(grid_values_t1, 2 * expected_values)



from enum import IntEnum
from unittest.mock import MagicMock

import pytest
import shapely.geometry as sg

from xugrid import meshkernel_utils as mku
from xugrid.constants import MissingOptionalModule

from . import requires_meshkernel

try:
    import meshkernel as mk

except ImportError:
    mk = MagicMock()
    mk.RefinementType = IntEnum(
        "RefinementType", ["WAVE_COURANT", "REFINEMENT_LEVELS", "RIDGE_DETECTION"]
    )


class Dummy(IntEnum):
    A = 1
    B = 2
    C = 3


def test_either_string_or_enum():
    assert (
        mku.either_string_or_enum("wave_courant", mk.RefinementType)
        == mk.RefinementType.WAVE_COURANT
    )
    assert (
        mku.either_string_or_enum("WAVE_COURANT", mk.RefinementType)
        == mk.RefinementType.WAVE_COURANT
    )
    assert (
        mku.either_string_or_enum("refinement_levels", mk.RefinementType)
        == mk.RefinementType.REFINEMENT_LEVELS
    )
    with pytest.raises(ValueError, match="Invalid option"):
        mku.either_string_or_enum("none", mk.RefinementType)
    with pytest.raises(TypeError, match="Option should be one of"):
        mku.either_string_or_enum(Dummy.A, mk.RefinementType)


@requires_meshkernel
def test_to_geometry_list():
    polygon = sg.Polygon(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]
    )
    actual = mku.to_geometry_list(polygon)
    assert isinstance(actual, mk.GeometryList)


def test_missing_optional_module():
    abc = MissingOptionalModule("abc")
    with pytest.raises(ImportError, match="abc is required for this functionality"):
        abc.attr



import pytest
import xarray as xr

import xugrid


def test_generate_disk():
    with pytest.raises(ValueError, match="partitions should be >= 3"):
        xugrid.data.generate_disk(2, 2)

    nodes, faces = xugrid.data.generate_disk(4, 1)
    assert nodes.shape == (5, 2)
    assert faces.shape == (4, 3)
    _, faces = xugrid.data.generate_disk(4, 2)
    assert faces.shape == (16, 3)


def test_adh_san_diego():
    ds = xugrid.data.adh_san_diego()
    assert isinstance(ds, xugrid.UgridDataset)
    ds = xugrid.data.adh_san_diego(xarray=True)
    assert isinstance(ds, xr.Dataset)


def test_disk():
    ds = xugrid.data.disk()
    assert isinstance(ds, xugrid.UgridDataset)


def test_xoxo():
    grid = xugrid.data.xoxo()
    assert isinstance(grid, xugrid.Ugrid2d)


def test_elevation_nl():
    ds = xugrid.data.elevation_nl()
    assert isinstance(ds, xugrid.UgridDataArray)
    ds = xugrid.data.elevation_nl(xarray=True)
    assert isinstance(ds, xr.Dataset)


def test_provinces_nl():
    import geopandas as gpd

    gdf = xugrid.data.provinces_nl()
    assert isinstance(gdf, gpd.GeoDataFrame)



"""
Copied from xarray.core.utils.py

The reason is that the content of xarray.core.utils are all private methods.
Hence, Xarray provides no guarantees on breaking changes.

Xarray is licensed under Apache License 2.0:
https://github.com/pydata/xarray/blob/main/LICENSE
"""

from __future__ import annotations

from collections.abc import Hashable, Mapping
from typing import Any, TypeVar, cast

from xugrid.ugrid.ugridbase import UgridType

T = TypeVar("T")


def is_dict_like(value: Any) -> Any:
    return hasattr(value, "keys") and hasattr(value, "__getitem__")


def either_dict_or_kwargs(
    pos_kwargs: Mapping[Any, T] | None,
    kw_kwargs: Mapping[str, T],
    func_name: str,
) -> Mapping[Hashable, T]:
    if pos_kwargs is None or pos_kwargs == {}:
        # Need an explicit cast to appease mypy due to invariance; see
        # https://github.com/python/mypy/issues/6228
        return cast(Mapping[Hashable, T], kw_kwargs)

    if not is_dict_like(pos_kwargs):
        raise ValueError(f"the first argument to .{func_name} must be a dictionary")
    if kw_kwargs:
        raise ValueError(
            f"cannot specify both keyword and positional arguments to .{func_name}"
        )
    return pos_kwargs


# EDIT: copied and simplified.
class UncachedAccessor:
    """Acts like a property, but on both classes and class instances

    This class is necessary because some tools (e.g. pydoc and sphinx)
    inspect classes for which property returns itself and not the
    accessor.
    """

    def __init__(self, accessor: type) -> None:
        self._accessor = accessor

    def __get__(self, obj: None | object, cls) -> Any:
        if obj is None:
            return self._accessor

        return self._accessor(obj)  # type: ignore  # assume it is a valid accessor!


def unique_grids(grids: list[UgridType]) -> list[UgridType]:
    uniques: list[UgridType] = []
    for grid in grids:
        for unique in uniques:
            if grid.equals(unique):
                break
        else:
            uniques.append(grid)
    return uniques



from typing import NamedTuple, Union

import numpy as np
from scipy.sparse import coo_matrix, csr_matrix

# import numpy.typing as npt

FloatDType = np.float64
IntDType = np.intp

# Requires numpy 1.21, not on conda yet...
# FloatArray = np.ndarray[FloatDType]
# IntArray = np.ndarray[IntDType]
# BoolArray = np.ndarray[np.bool_]

FloatArray = np.ndarray
IntArray = np.ndarray
BoolArray = np.ndarray
# Pygeos collections:
PointArray = np.ndarray
LineArray = np.ndarray
PolygonArray = np.ndarray
SparseMatrix = Union[coo_matrix, csr_matrix]

# Internally we always use a fill value of -1. This ensures we can always index
# with the fill value as well, since any array will have at least size 1.
FILL_VALUE = -1


class Point(NamedTuple):
    x: float
    y: float


class Vector(NamedTuple):
    x: float
    y: float


# Spatial coordinate epsilon for floating point comparison
# Assuming world coordinates in meters: 40 000 m along equator:
# 40 000 000 = 4e7 mm
# np.spacing(4e7) == 7.45E-9 ~= 1E-8
X_EPSILON = 1.0e-8
X_OFFSET = 1.0e-8
T_OFFSET = 1.0e-6


class MissingOptionalModule:
    """Presents a clear error for optional modules."""

    def __init__(self, name):
        self.name = name

    def __getattr__(self, name):
        raise ImportError(f"{self.name} is required for this functionality")



"""
The full content is this module is copied from xarray.plot.utils.py. Unused
parts are removed, so this module only contains about a third of the original's
content.

Additionally, _easy_facetgrid has been copied from xarray.plot.facetgrid.

The reason is that these functions are all private methods. Hence, Xarray
provides no guarantees on breaking changes.

We heavily discourage editing this file. Any update should only consist of
copying updated parts of the xarray module.

Some minor edits are necessary. These are marked by an comment with "EDIT:".

Xarray is licensed under Apache License 2.0:
https://github.com/pydata/xarray/blob/main/LICENSE
"""

from __future__ import annotations

import importlib
import textwrap
from collections.abc import Callable, Hashable, Iterable, Mapping
from datetime import datetime
from inspect import getfullargspec
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union

import numpy as np
from xarray import DataArray, Dataset
from xarray.plot.facetgrid import FacetGrid

# EDIT: added these here
AspectOptions = Union[Literal["auto", "equal"], float, None]
ScaleOptions = Literal["linear", "symlog", "log", "logit", None]
T_DataArrayOrSet = TypeVar("T_DataArrayOrSet", bound=Union["Dataset", "DataArray"])
nc_time_axis_available = importlib.util.find_spec("nc_time_axis") is not None


try:
    import cftime
except ImportError:
    cftime = None


if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from numpy.typing import ArrayLike

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        plt: Any = None  # type: ignore

ROBUST_PERCENTILE = 2.0


__all__ = [
    "_add_colorbar",
    "_ensure_plottable",
    "_process_cmap_cbar_kwargs",
    "_update_axes",
    "get_axis",
    "label_from_attrs",
    "_easy_facetgrid",
]


def _determine_extend(calc_data, vmin, vmax):
    extend_min = calc_data.min() < vmin
    extend_max = calc_data.max() > vmax
    if extend_min and extend_max:
        return "both"
    elif extend_min:
        return "min"
    elif extend_max:
        return "max"
    else:
        return "neither"


def _build_discrete_cmap(cmap, levels, extend, filled):
    """Build a discrete colormap and normalization of the data."""
    import matplotlib as mpl

    if len(levels) == 1:
        levels = [levels[0], levels[0]]

    if not filled:
        # non-filled contour plots
        extend = "max"

    if extend == "both":
        ext_n = 2
    elif extend in ["min", "max"]:
        ext_n = 1
    else:
        ext_n = 0

    n_colors = len(levels) + ext_n - 1
    pal = _color_palette(cmap, n_colors)

    new_cmap, cnorm = mpl.colors.from_levels_and_colors(levels, pal, extend=extend)
    # copy the old cmap name, for easier testing
    new_cmap.name = getattr(cmap, "name", cmap)

    # copy colors to use for bad, under, and over values in case they have been
    # set to non-default values
    try:
        # matplotlib<3.2 only uses bad color for masked values
        bad = cmap(np.ma.masked_invalid([np.nan]))[0]
    except TypeError:
        # cmap was a str or list rather than a color-map object, so there are
        # no bad, under or over values to check or copy
        pass
    else:
        under = cmap(-np.inf)
        over = cmap(np.inf)

        new_cmap.set_bad(bad)

        # Only update under and over if they were explicitly changed by the user
        # (i.e. are different from the lowest or highest values in cmap). Otherwise
        # leave unchanged so new_cmap uses its default values (its own lowest and
        # highest values).
        if under != cmap(0):
            new_cmap.set_under(under)
        if over != cmap(cmap.N - 1):
            new_cmap.set_over(over)

    return new_cmap, cnorm


def _color_palette(cmap, n_colors):
    import matplotlib.pyplot as plt
    from matplotlib.colors import ListedColormap

    colors_i = np.linspace(0, 1.0, n_colors)
    if isinstance(cmap, (list, tuple)):
        # we have a list of colors
        cmap = ListedColormap(cmap, N=n_colors)
        pal = cmap(colors_i)
    elif isinstance(cmap, str):
        # we have some sort of named palette
        try:
            # is this a matplotlib cmap?
            cmap = plt.get_cmap(cmap)
            pal = cmap(colors_i)
        except ValueError:
            # ValueError happens when mpl doesn't like a colormap, try seaborn
            try:
                from seaborn import color_palette

                pal = color_palette(cmap, n_colors=n_colors)
            except (ValueError, ImportError):
                # or maybe we just got a single color as a string
                cmap = ListedColormap([cmap], N=n_colors)
                pal = cmap(colors_i)
    else:
        # cmap better be a LinearSegmentedColormap (e.g. viridis)
        pal = cmap(colors_i)

    return pal


# _determine_cmap_params is adapted from Seaborn:
# https://github.com/mwaskom/seaborn/blob/v0.6/seaborn/matrix.py#L158
# Used under the terms of Seaborn's license, see licenses/SEABORN_LICENSE.


def _determine_cmap_params(
    plot_data,
    vmin=None,
    vmax=None,
    cmap=None,
    center=None,
    robust=False,
    extend=None,
    levels=None,
    filled=True,
    norm=None,
    _is_facetgrid=False,
):
    """
    Use some heuristics to set good defaults for colorbar and range.

    Parameters
    ----------
    plot_data : Numpy array
        Doesn't handle xarray objects

    Returns
    -------
    cmap_params : dict
        Use depends on the type of the plotting function
    """
    import matplotlib as mpl

    if isinstance(levels, Iterable):
        levels = sorted(levels)

    calc_data = np.ravel(plot_data[np.isfinite(plot_data)])

    # Handle all-NaN input data gracefully
    if calc_data.size == 0:
        # Arbitrary default for when all values are NaN
        calc_data = np.array(0.0)

    # Setting center=False prevents a divergent cmap
    possibly_divergent = center is not False

    # Set center to 0 so math below makes sense but remember its state
    center_is_none = False
    if center is None:
        center = 0
        center_is_none = True

    # Setting both vmin and vmax prevents a divergent cmap
    if (vmin is not None) and (vmax is not None):
        possibly_divergent = False

    # Setting vmin or vmax implies linspaced levels
    user_minmax = (vmin is not None) or (vmax is not None)

    # vlim might be computed below
    vlim = None

    # save state; needed later
    vmin_was_none = vmin is None
    vmax_was_none = vmax is None

    if vmin is None:
        if robust:
            vmin = np.percentile(calc_data, ROBUST_PERCENTILE)
        else:
            vmin = calc_data.min()
    elif possibly_divergent:
        vlim = abs(vmin - center)

    if vmax is None:
        if robust:
            vmax = np.percentile(calc_data, 100 - ROBUST_PERCENTILE)
        else:
            vmax = calc_data.max()
    elif possibly_divergent:
        vlim = abs(vmax - center)

    if possibly_divergent:
        levels_are_divergent = (
            isinstance(levels, Iterable) and levels[0] * levels[-1] < 0
        )
        # kwargs not specific about divergent or not: infer defaults from data
        divergent = (
            ((vmin < 0) and (vmax > 0)) or not center_is_none or levels_are_divergent
        )
    else:
        divergent = False

    # A divergent map should be symmetric around the center value
    if divergent:
        if vlim is None:
            vlim = max(abs(vmin - center), abs(vmax - center))
        vmin, vmax = -vlim, vlim

    # Now add in the centering value and set the limits
    vmin += center
    vmax += center

    # now check norm and harmonize with vmin, vmax
    if norm is not None:
        if norm.vmin is None:
            norm.vmin = vmin
        else:
            if not vmin_was_none and vmin != norm.vmin:
                raise ValueError("Cannot supply vmin and a norm with a different vmin.")
            vmin = norm.vmin

        if norm.vmax is None:
            norm.vmax = vmax
        else:
            if not vmax_was_none and vmax != norm.vmax:
                raise ValueError("Cannot supply vmax and a norm with a different vmax.")
            vmax = norm.vmax

    # if BoundaryNorm, then set levels
    if isinstance(norm, mpl.colors.BoundaryNorm):
        levels = norm.boundaries

    # Choose default colormaps if not provided
    if cmap is None:
        # EDIT: replaced OPTIONS
        if divergent:
            cmap = "RdBu_r"
        else:
            cmap = "viridis"

    # Handle discrete levels
    if levels is not None:
        # EDIT: replaced is_scalar(levels) by np.issubdtype(type(levels), np.integer)
        # xarray uses a relatelively complicated is_scalar function here. For
        # the purposes of plotting, checking for an integer should do.
        if np.issubdtype(type(levels), np.integer):
            if user_minmax:
                levels = np.linspace(vmin, vmax, levels)
            elif levels == 1:
                levels = np.asarray([(vmin + vmax) / 2])
            else:
                # N in MaxNLocator refers to bins, not ticks
                ticker = mpl.ticker.MaxNLocator(levels - 1)
                levels = ticker.tick_values(vmin, vmax)
        vmin, vmax = levels[0], levels[-1]

    # GH3734
    if vmin == vmax:
        vmin, vmax = mpl.ticker.LinearLocator(2).tick_values(vmin, vmax)

    if extend is None:
        extend = _determine_extend(calc_data, vmin, vmax)

    if (levels is not None) and (not isinstance(norm, mpl.colors.BoundaryNorm)):
        cmap, newnorm = _build_discrete_cmap(cmap, levels, extend, filled)
        norm = newnorm if norm is None else norm

    # vmin & vmax needs to be None if norm is passed
    # TODO: always return a norm with vmin and vmax
    if norm is not None:
        vmin = None
        vmax = None

    return {
        "vmin": vmin,
        "vmax": vmax,
        "cmap": cmap,
        "extend": extend,
        "levels": levels,
        "norm": norm,
    }


def get_axis(
    figsize: Iterable[float] | None = None,
    size: float | None = None,
    aspect: AspectOptions = None,
    ax: Axes | None = None,
    **subplot_kws: Any,
) -> Axes:
    try:
        import matplotlib as mpl
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("matplotlib is required for plot.utils.get_axis")

    if figsize is not None:
        if ax is not None:
            raise ValueError("cannot provide both `figsize` and `ax` arguments")
        if size is not None:
            raise ValueError("cannot provide both `figsize` and `size` arguments")
        _, ax = plt.subplots(figsize=figsize, subplot_kw=subplot_kws)
        return ax

    if size is not None:
        if ax is not None:
            raise ValueError("cannot provide both `size` and `ax` arguments")
        if aspect is None or aspect == "auto":
            width, height = mpl.rcParams["figure.figsize"]
            faspect = width / height
        elif aspect == "equal":
            faspect = 1
        else:
            faspect = aspect
        figsize = (size * faspect, size)
        _, ax = plt.subplots(figsize=figsize, subplot_kw=subplot_kws)
        return ax

    if aspect is not None:
        raise ValueError("cannot provide `aspect` argument without `size`")

    if subplot_kws and ax is not None:
        raise ValueError("cannot use subplot_kws with existing ax")

    if ax is None:
        ax = _maybe_gca(**subplot_kws)

    return ax


def _maybe_gca(**subplot_kws: Any) -> Axes:
    import matplotlib.pyplot as plt

    # can call gcf unconditionally: either it exists or would be created by plt.axes
    f = plt.gcf()

    # only call gca if an active axes exists
    if f.axes:
        # can not pass kwargs to active axes
        return plt.gca()

    return plt.axes(**subplot_kws)


def _get_units_from_attrs(da: DataArray) -> str:
    """Extract and format the unit/units from a attributes."""
    # EDIT: removed pint support for now.
    # pint_array_type = DuckArrayModule("pint").type
    units = " [{}]"
    # if isinstance(da.data, pint_array_type):
    #    return units.format(str(da.data.units))
    if "units" in da.attrs:
        return units.format(da.attrs["units"])
    if "unit" in da.attrs:
        return units.format(da.attrs["unit"])
    return ""


def label_from_attrs(da: DataArray | None, extra: str = "") -> str:
    """
    Make informative labels if variable metadata (attrs) follows CF
    conventions.
    """
    if da is None:
        return ""

    name: str = "{}"
    if "long_name" in da.attrs:
        name = name.format(da.attrs["long_name"])
    elif "standard_name" in da.attrs:
        name = name.format(da.attrs["standard_name"])
    elif da.name is not None:
        name = name.format(da.name)
    else:
        name = ""

    units = _get_units_from_attrs(da)

    # Treat `name` differently if it's a latex sequence
    if name.startswith("$") and (name.count("$") % 2 == 0):
        return "$\n$".join(
            textwrap.wrap(name + extra + units, 60, break_long_words=False)
        )
    else:
        return "\n".join(textwrap.wrap(name + extra + units, 30))


def _valid_other_type(
    x: ArrayLike, types: type[object] | tuple[type[object], ...]
) -> bool:
    """Do all elements of x have a type from types?"""
    return all(isinstance(el, types) for el in np.ravel(x))


def _valid_numpy_subdtype(x, numpy_types):
    """Is any dtype from numpy_types superior to the dtype of x?"""
    # If any of the types given in numpy_types is understood as numpy.generic,
    # all possible x will be considered valid.  This is probably unwanted.
    for t in numpy_types:
        assert not np.issubdtype(np.generic, t)

    return any(np.issubdtype(x.dtype, t) for t in numpy_types)


def _ensure_plottable(*args) -> None:
    """
    Raise exception if there is anything in args that can't be plotted on an
    axis by matplotlib.
    """
    numpy_types: tuple[type[object], ...] = (
        np.floating,
        np.integer,
        np.timedelta64,
        np.datetime64,
        np.bool_,
        np.str_,
    )
    other_types: tuple[type[object], ...] = (datetime,)
    cftime_datetime_types: tuple[type[object], ...] = (
        () if cftime is None else (cftime.datetime,)
    )
    other_types += cftime_datetime_types

    for x in args:
        if not (
            _valid_numpy_subdtype(np.asarray(x), numpy_types)
            or _valid_other_type(np.asarray(x), other_types)
        ):
            raise TypeError(
                "Plotting requires coordinates to be numeric, boolean, "
                "or dates of type numpy.datetime64, "
                "datetime.datetime, cftime.datetime or "
                f"pandas.Interval. Received data of type {np.asarray(x).dtype} instead."
            )
        if _valid_other_type(np.asarray(x), cftime_datetime_types):
            if nc_time_axis_available:
                # Register cftime datetypes to matplotlib.units.registry,
                # otherwise matplotlib will raise an error:
                import nc_time_axis  # noqa: F401
            else:
                raise ImportError(
                    "Plotting of arrays of cftime.datetime "
                    "objects or arrays indexed by "
                    "cftime.datetime objects requires the "
                    "optional `nc-time-axis` (v1.2.0 or later) "
                    "package."
                )


def _add_colorbar(primitive, ax, cbar_ax, cbar_kwargs, cmap_params):
    cbar_kwargs.setdefault("extend", cmap_params["extend"])
    if cbar_ax is None:
        cbar_kwargs.setdefault("ax", ax)
    else:
        cbar_kwargs.setdefault("cax", cbar_ax)

    # dont pass extend as kwarg if it is in the mappable
    if hasattr(primitive, "extend"):
        cbar_kwargs.pop("extend")

    fig = ax.get_figure()
    cbar = fig.colorbar(primitive, **cbar_kwargs)

    return cbar


def _update_axes(
    ax: Axes,
    xincrease: bool | None,
    yincrease: bool | None,
    xscale: ScaleOptions = None,
    yscale: ScaleOptions = None,
    xticks: ArrayLike | None = None,
    yticks: ArrayLike | None = None,
    xlim: tuple[float, float] | None = None,
    ylim: tuple[float, float] | None = None,
) -> None:
    """Update axes with provided parameters"""
    if xincrease is None:
        pass
    elif xincrease and ax.xaxis_inverted():
        ax.invert_xaxis()
    elif not xincrease and not ax.xaxis_inverted():
        ax.invert_xaxis()

    if yincrease is None:
        pass
    elif yincrease and ax.yaxis_inverted():
        ax.invert_yaxis()
    elif not yincrease and not ax.yaxis_inverted():
        ax.invert_yaxis()

    # The default xscale, yscale needs to be None.
    # If we set a scale it resets the axes formatters,
    # This means that set_xscale('linear') on a datetime axis
    # will remove the date labels. So only set the scale when explicitly
    # asked to. https://github.com/matplotlib/matplotlib/issues/8740
    if xscale is not None:
        ax.set_xscale(xscale)
    if yscale is not None:
        ax.set_yscale(yscale)

    if xticks is not None:
        ax.set_xticks(xticks)
    if yticks is not None:
        ax.set_yticks(yticks)

    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)


def _process_cmap_cbar_kwargs(
    func,
    data,
    cmap=None,
    colors=None,
    cbar_kwargs: Iterable[tuple[str, Any]] | Mapping[str, Any] | None = None,
    levels=None,
    _is_facetgrid=False,
    **kwargs,
) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Parameters
    ----------
    func : plotting function
    data : ndarray,
        Data values

    Returns
    -------
    cmap_params : dict
    cbar_kwargs : dict
    """
    if func.__name__ == "surface":
        # Leave user to specify cmap settings for surface plots
        kwargs["cmap"] = cmap
        return {
            k: kwargs.get(k, None)
            for k in ["vmin", "vmax", "cmap", "extend", "levels", "norm"]
        }, {}

    cbar_kwargs = {} if cbar_kwargs is None else dict(cbar_kwargs)

    if "contour" in func.__name__ and levels is None:
        levels = 7  # this is the matplotlib default

    # colors is mutually exclusive with cmap
    if cmap and colors:
        raise ValueError("Can't specify both cmap and colors.")

    # colors is only valid when levels is supplied or the plot is of type
    # contour or contourf
    if colors and (("contour" not in func.__name__) and (levels is None)):
        raise ValueError("Can only specify colors with contour or levels")

    # we should not be getting a list of colors in cmap anymore
    # is there a better way to do this test?
    if isinstance(cmap, (list, tuple)):
        raise ValueError(
            "Specifying a list of colors in cmap is deprecated. "
            "Use colors keyword instead."
        )

    cmap_kwargs = {
        "plot_data": data,
        "levels": levels,
        "cmap": colors if colors else cmap,
        "filled": func.__name__ != "contour",
    }

    cmap_args = getfullargspec(_determine_cmap_params).args
    cmap_kwargs.update((a, kwargs[a]) for a in cmap_args if a in kwargs)
    if not _is_facetgrid:
        cmap_params = _determine_cmap_params(**cmap_kwargs)
    else:
        cmap_params = {
            k: cmap_kwargs[k]
            for k in ["vmin", "vmax", "cmap", "extend", "levels", "norm"]
        }

    return cmap_params, cbar_kwargs


def _easy_facetgrid(
    data: T_DataArrayOrSet,
    plotfunc: Callable,
    kind: Literal["line", "dataarray", "dataset", "plot1d"],
    x: Hashable | None = None,
    y: Hashable | None = None,
    row: Hashable | None = None,
    col: Hashable | None = None,
    col_wrap: int | None = None,
    sharex: bool = True,
    sharey: bool = True,
    aspect: float | None = None,
    size: float | None = None,
    subplot_kws: dict[str, Any] | None = None,
    ax: Axes | None = None,
    figsize: Iterable[float] | None = None,
    **kwargs: Any,
) -> FacetGrid[T_DataArrayOrSet]:
    """
    Convenience method to call xarray.plot.FacetGrid from 2d plotting methods

    kwargs are the arguments to 2d plotting method
    """  # noqa: D401
    if ax is not None:
        raise ValueError("Can't use axes when making faceted plots.")
    if aspect is None:
        aspect = 1
    if size is None:
        size = 3
    elif figsize is not None:
        raise ValueError("cannot provide both `figsize` and `size` arguments")
    if kwargs.get("z") is not None:
        # 3d plots doesn't support sharex, sharey, reset to mpl defaults:
        sharex = False
        sharey = False

    g = FacetGrid(
        data=data,
        col=col,
        row=row,
        col_wrap=col_wrap,
        sharex=sharex,
        sharey=sharey,
        figsize=figsize,
        aspect=aspect,
        size=size,
        subplot_kws=subplot_kws,
    )

    if kind == "line":
        return g.map_dataarray_line(plotfunc, x, y, **kwargs)

    if kind == "dataarray":
        return g.map_dataarray(plotfunc, x, y, **kwargs)

    if kind == "plot1d":
        return g.map_plot1d(plotfunc, x, y, **kwargs)

    if kind == "dataset":
        return g.map_dataset(plotfunc, x, y, **kwargs)

    raise ValueError(
        f"kind must be one of `line`, `dataarray`, `dataset` or `plot1d`, got {kind}"
    )



"""
Custom Sparse Matrix utilities.

Numba cannot deal with scipy.sparse objects directly. The data
structures are mostly a collection of numpy arrays, which can
be neatly represented by (typed) namedtuples, which numba accepts.
"""

from typing import NamedTuple

import numba
import numpy as np
from scipy import sparse

from xugrid.constants import FloatArray, IntArray, IntDType


class MatrixCOO(NamedTuple):
    """
    A sparse matrix in COOrdinate format, also known as the "ijv" or "triplet"
    format. More or less matches the scipy.sparse.coo_matrix.

    NamedTuple for easy ingestion by numba.

    Parameters
    ----------
    data: np.ndarray of floats
        The regridding weights.
    row: np.ndarray of integers
        The linear index into the source grid.
    col: np.ndarray of integers
        The linear index into the target grid.
    n: int
        The number of rows.
    m: int
        The number of columns.
    nnz: int
        The number of non-zero values.
    """

    data: FloatArray
    row: IntArray
    col: IntArray
    n: int
    m: int
    nnz: int

    @staticmethod
    def from_triplet(row, col, data, n=None, m=None) -> "MatrixCOO":
        if n is None:
            n = row.max() + 1
        if m is None:
            m = col.max() + 1
        nnz = data.size
        return MatrixCOO(data, row, col, n, m, nnz)

    def to_csr(self) -> "MatrixCSR":
        """
        Convert COO matrix to CSR matrix.

        Assumes the COO matrix indices are already sorted by row number!
        """
        i = np.cumsum(np.bincount(self.row, minlength=self.n))
        indptr = np.empty(i.size + 1, dtype=IntDType)
        indptr[0] = 0
        indptr[1:] = i
        return MatrixCSR(
            self.data,
            self.col,
            indptr,
            self.n,
            self.m,
            self.nnz,
        )


class MatrixCSR(NamedTuple):
    """
    Compressed Sparse Row matrix. The row indices are compressed; all values
    must therefore be sorted by row number. More or less matches the
    scipy.sparse.csr_matrix.

    NamedTuple for easy ingestion by numba.

    Parameters
    ----------
    data: np.ndarray of floats
        The values of the matrix.
    indices: np.ndarray of integers
        The column numbers of the CSR format.
    indptr: inp.ndarray of integers
        The row index CSR pointer array.
        Values for row i (target index i) are stored in:
        indices[indptr[i]: indptr[i + 1]]
    n: int
        The number of rows.
    m: int
        The number of columns.
    nnz: int
        The number of non-zero values.
    """

    data: FloatArray
    indices: IntArray
    indptr: IntArray
    n: int
    m: int
    nnz: int

    @staticmethod
    def from_csr_matrix(A: sparse.csr_matrix) -> "MatrixCSR":
        n, m = A.shape
        return MatrixCSR(A.data, A.indices, A.indptr, n, m, A.nnz)

    @staticmethod
    def from_triplet(
        row,
        col,
        data,
        n=None,
        m=None,
    ) -> "MatrixCSR":
        return MatrixCOO.from_triplet(row, col, data, n, m).to_csr()

    def to_coo(self) -> MatrixCOO:
        """
        Convert CSR matrix to COO matrix.

        Expand the indtpr to full row numbers.
        """
        n_repeat = np.diff(self.indptr)
        row = np.repeat(np.arange(self.n), n_repeat)
        return MatrixCOO(self.data, row, self.indices, self.n, self.m, self.nnz)


@numba.njit(inline="always")
def nzrange(A: MatrixCSR, row: int) -> range:
    """Return the non-zero indices of a single row."""
    start = A.indptr[row]
    end = A.indptr[row + 1]
    return range(start, end)


@numba.njit(inline="always")
def row_slice(A, row: int) -> slice:
    """Return the indices or data slice of a single row."""
    start = A.indptr[row]
    end = A.indptr[row + 1]
    return slice(start, end)


@numba.njit(inline="always")
def columns_and_values(A, slice):
    return zip(A.indices[slice], A.data[slice])
